import fcntl
import os
import logging
import pty
import re
import select
import subprocess
import errno
from socket import socket, AF_INET, SOCK_STREAM
from contextlib import closing

import attr

from ..step import step

re_vt100 = re.compile(r"(\x1b\[|\x9b)[^@-_a-z]*[@-_a-z]|\x1b[@-_a-z]")

def get_free_port():
    """Helper function to always return an unused port."""
    with closing(socket(AF_INET, SOCK_STREAM)) as s:
        s.bind(('', 0))
        return s.getsockname()[1]


def get_user():
    """Get the username of the current user."""
    user = os.environ.get("USER")
    if user:
        return user
    import pwd
    return pwd.getpwuid(os.getuid())[0]

def set_nonblocking(fd):
    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

@attr.s
class ProcessWrapper:
    callbacks = attr.ib(default=attr.Factory(list))
    loglevel = logging.INFO

    @step(args=['command'], result=True, tag='process')
    def check_output(self, command, *, print_on_silent_log=False, input=None, stdin=None): # pylint: disable=redefined-builtin
        """Run a command and supply the output to callback functions"""
        logger = logging.getLogger("Process")
        res = []
        mfd, sfd = pty.openpty()
        set_nonblocking(mfd)

        kwargs = {}

        stdin_r = None
        stdin_w = None
        if input is not None:
            stdin_r, stdin_w = os.pipe()
            kwargs['stdin'] = stdin_r
            set_nonblocking(stdin_w)
        elif stdin is not None:
            kwargs['stdin'] = stdin

        process = subprocess.Popen(command, stderr=sfd,
                                   stdout=sfd, bufsize=0, **kwargs)

        logger.log(ProcessWrapper.loglevel, "[%d] command: %s", process.pid, " ".join(command))

        # do not register/unregister already registered print_callback
        if ProcessWrapper.print_callback in self.callbacks:
            print_on_silent_log = False

        if print_on_silent_log and logger.getEffectiveLevel() > ProcessWrapper.loglevel:
            self.enable_print()

        if stdin_r is not None:
            os.close(stdin_r)

        # close sfd so we notice when the child is gone
        os.close(sfd)
        # get a file object from the fd
        buf = b""
        read_fds = [mfd]
        write_fds = []
        if stdin_w is not None:
            write_fds.append(stdin_w)

        while True:
            ready_r, ready_w, _ = select.select(read_fds, write_fds, [], 0.1)

            if mfd in ready_r:
                raw = None

                try:
                    raw = os.read(mfd, 4096)
                except BlockingIOError:
                    pass
                except OSError as e:
                    if e.errno == errno.EIO:
                        break
                    raise

                if raw:
                    buf += raw
                    *parts, buf = buf.split(b'\r')
                    res.extend(parts)
                    for part in parts:
                        for callback in self.callbacks:
                            callback(part, process)

            if stdin_w in ready_w:
                if input:
                    amt = 0
                    try:
                        amt = os.write(stdin_w, input)
                    except BlockingIOError:
                        pass
                    except OSError as e:
                        if e.errno == errno.EIO:
                            break
                        raise
                    input = input[amt:]

                if not input:
                    write_fds.remove(stdin_w)
                    os.close(stdin_w)
                    stdin_w = None

            process.poll()
            if process.returncode is not None:
                break

        if stdin_w is not None:
            os.close(stdin_w)

        os.close(mfd)
        process.wait()
        if buf:
            # process incomplete line
            res.append(buf)
            if buf[-1] != b'\n':
                buf += b'\n'
            for callback in self.callbacks:
                callback(buf, process)

        if print_on_silent_log and logger.getEffectiveLevel() > ProcessWrapper.loglevel:
            self.disable_print()

        if process.returncode != 0:
            raise subprocess.CalledProcessError(process.returncode,
                                                command,
                                                output=b'\r'.join(res))
        # this converts '\r\n' to '\n' to be more compatible to the behaviour
        # of the normal subprocess module
        return b'\n'.join([r.strip(b'\n') for r in res])

    def register(self, callback):
        """Register a callback with the ProcessWrapper"""
        if callback in self.callbacks:
            return
        self.callbacks.append(callback)

    def unregister(self, callback):
        """Unregister a callback with the ProcessWrapper"""
        if callback not in self.callbacks:
            return
        self.callbacks.remove(callback)

    @staticmethod
    def log_callback(message, process):
        """Logs process output message along with its pid."""
        logger = logging.getLogger("Process")
        message = message.decode(encoding="utf-8", errors="replace").strip("\n")
        if message:
            logger.log(ProcessWrapper.loglevel, "[%d] %s", process.pid, message)

    @staticmethod
    def print_callback(message, _):
        """Prints process output message."""
        message = message.decode(encoding="utf-8", errors="replace")
        print(f"\r{message}", end='')

    def enable_logging(self):
        """Enables process output to the logging interface."""
        self.register(ProcessWrapper.log_callback)

    def disable_logging(self):
        """Disables process output logging."""
        self.unregister(ProcessWrapper.log_callback)

    def enable_print(self):
        """Enables process output to print."""
        self.register(ProcessWrapper.print_callback)

    def disable_print(self):
        """Disables process output printing."""
        self.unregister(ProcessWrapper.print_callback)


processwrapper = ProcessWrapper()



from importlib import import_module
import attr

from ..exceptions import InvalidConfigError
from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..util.proxy import proxymanager
from .common import Driver
from .exception import ExecutionError
from ..step import step


@target_factory.reg_driver
@attr.s(eq=False)
class OneWirePIODriver(Driver, DigitalOutputProtocol):
    bindings = {
        "port": "OneWirePIO",
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._module = import_module("onewire")
        self._onewire = None
        if "PIO" not in self.port.path:
            raise InvalidConfigError(
                f"Invalid OneWire path {self.port.path} (needs to be in the form of ??.????????????/PIO.?)"  # pylint: disable=line-too-long
            )

    def on_activate(self):
        # we can only forward if the backend knows which port to use
        host, port = proxymanager.get_host_and_port(self.port)
        self._onewire = self._module.Onewire(f"{host}:{port}")

    def on_deactivate(self):
        self._onewire = None

    @Driver.check_active
    @step(args=["status"])
    def set(self, status):
        if self.port.invert:
            status = not status
        if status:
            self._onewire.set(self.port.path, "1")
        else:
            self._onewire.set(self.port.path, "0")

    @Driver.check_active
    @step(result=["True"])
    def get(self):
        path = self.port.path.replace("PIO", "sensed")
        status = self._onewire.get(path)
        if status is None:
            raise ExecutionError(f"Failed to get OneWire value for {path}")
        if status not in ["0", "1"]:
            raise ExecutionError(f"Invalid OneWire value ({repr(status)})")
        status = status == "1"
        if self.port.invert:
            status = not status
        return status



#!/usr/bin/env python

import base64
import ctypes
import fcntl
import os
import pickle
import socket
import struct
import subprocess
from pathlib import Path

CLONE_NEWNS = 0x00020000  # Mount namespace
CLONE_NEWUSER = 0x10000000  # User namespace
CLONE_NEWNET = 0x40000000  # Network namespace

libc = ctypes.CDLL("libc.so.6", use_errno=True)

libc.unshare.argtypes = [ctypes.c_int]
libc.unshare.restype = ctypes.c_int

libc.setns.argtypes = [ctypes.c_int, ctypes.c_int]
libc.setns.restype = ctypes.c_int

socket_table = {}


def b2s(b):
    return base64.b85encode(b).decode('ascii')


def s2b(s):
    return base64.b85decode(s.encode('ascii'))


def py2s(o):
    return b2s(pickle.dumps(o))


def s2py(s):
    return pickle.loads(s2b(s))


def unshare(flags):
    ret = libc.unshare(flags)
    if ret != 0:
        err = ctypes.get_errno()
        raise OSError(err, os.strerror(err))
    return ret


def setns(fd, nstype=0):
    ret = libc.setns(fd, nstype)
    if ret != 0:
        err = ctypes.get_errno()
        raise OSError(err, os.strerror(err))
    return ret


unshared = False


def handle_unshare():
    global unshared
    assert not unshared

    uid = os.getuid()
    gid = os.getgid()

    unshare(CLONE_NEWUSER)
    unshare(CLONE_NEWNET)

    uidmap = Path("/proc/self/uid_map")
    uidmap.write_text(f"0 {uid} 1")
    setgroups = Path("/proc/self/setgroups")
    setgroups.write_text("deny")
    gidmap = Path("/proc/self/gid_map")
    gidmap.write_text(f"0 {gid} 1")

    # mount again from inside the netns, so that the correct devices are visible
    # subprocess.check_call(["mount", "-t", "sysfs", "sysfs", "/sys"])
    # subprocess.check_call(['mount', '-t', 'devtmpfs', 'devtmpfs', '/dev'])
    # subprocess.check_call(['mount', '-t', 'proc', 'proc', '/proc'])

    subprocess.check_call(["ip", "link", "set", "lo", "up"])

    unshared = True

    return os.getpid()


def handle_create_tun(*, address=None):
    with os.fdopen(os.open("/dev/net/tun", os.O_RDWR)) as dev_tun:
        TUNSETIFF = 0x400454CA
        IFF_TAP = 0x0002
        IFF_NO_PI = 0x1000
        ifr = struct.pack("16sH22s", b"tap0", IFF_TAP | IFF_NO_PI, b"\x00" * 22)
        fcntl.ioctl(dev_tun, TUNSETIFF, ifr)

        subprocess.run(["ip", "link", "set", "up", "tap0"], check=True)
        if address:
            subprocess.run(["ip", "link", "set", "address", address, "dev", "tap0"], check=True)

        return ("", os.fdopen(os.dup(dev_tun.fileno())))


def handle_socket(*args, **kwargs):
    try:
        # The socket creation parameters are constant integers defined by the
        # kernel. Since it's not possible to pass a file descriptor from a
        # remote agent, there is no danger of these constants being out of sync
        # so pass them directly instead of using strings.
        s = socket.socket(*args, **kwargs)
        socket_table[id(s)] = s
        return ({"id": id(s)}, s.dup())
    except OSError as e:
        return ({"error": [e.errno, e.strerror]}, None)


def handle_socket_close(sockid):
    if sockid in socket_table:
        socket_table[sockid].close()
        del socket_table[sockid]


def handle_socket_dup(sockid):
    try:
        s = socket_table[sockid].dup()
        socket_table[id(s)] = s
        return ({"id": id(s)}, s.dup())
    except OSError as e:
        return ({"error": [e.errno, e.strerror]}, None)


def handle_socket_call(sockid, func, args, kwargs):
    try:
        s = socket_table[sockid]
        ret = getattr(s, func)(*s2py(args), **s2py(kwargs))
        return (0, py2s(ret))
    except OSError as e:
        return ([e.errno, e.strerror], None)


def handle_list_sockets():
    return list(socket_table.keys())


def handle_get_links():
    import json

    output = subprocess.check_output(["ip", "-j", "link"])
    return json.loads(output)


def handle_get_prefix():
    """Returns the command prefix to use to execute commands in the namespace"""
    return ["nsenter", "-t", str(os.getpid()), "-U", "-n", "-m", "--preserve-credentials"]


def handle_get_pid():
    """Returns the PID of the namespace"""
    return os.getpid()


def handle_get_intf():
    return "tap0"


def handle_getaddrinfo(*args, **kwargs):
    """
    Call getaddrinfo in the namespace.

    """
    result = []
    try:
        for info in socket.getaddrinfo(*args, **kwargs):
            result.append(info)

        return (0, result)
    except OSError as e:
        return ((e.errno, e.strerror), result)


methods = {
    "unshare": handle_unshare,
    "create_tun": handle_create_tun,
    "create_socket": handle_socket,
    "get_links": handle_get_links,
    "get_prefix": handle_get_prefix,
    "get_pid": handle_get_pid,
    "get_intf": handle_get_intf,
    "socket_call": handle_socket_call,
    "socket_close": handle_socket_close,
    "socket_dup": handle_socket_dup,
    "list_sockets": handle_list_sockets,
    "getaddrinfo": handle_getaddrinfo,
}



import requests

# Driver has been tested with:
# * Gude Expert Power Control 8031()
# * Gude Expert Power Control 87-1210-18
#   This device needs to be used in 'Basic Compatible' mode for HTTP GET
#   to be usable. Do not turn on session authentication.

# HTTP-GET API is defined in the Gude EPC-HTTP-Interface specification:
# http://wiki.gude.info/EPC_HTTP_Interface
#
# The `components=<N>` parameter defines which status information are
# included into the returned JSON.
# * `components=0` happily returns an empty response but still switches the
#   outputs as requested.
# * `components=1` only includes the output's state into the JSON.

PORT = 80


def power_set(host, port, index, value):
    index = int(index)
    assert 1 <= index <= 20
    # access the web interface...
    value = 1 if value else 0
    r = requests.get(f"http://{host}:{port}/status.json?components=0&cmd=1&p={index}&s={value}")
    r.raise_for_status()


def power_get(host, port, index):
    index = int(index)
    assert 1 <= index <= 20

    # get the component status
    r = requests.get(f"http://{host}:{port}/status.json?components=1")
    r.raise_for_status()

    state = r.json()["outputs"][index - 1]["state"]

    return state



import pytest

from labgrid.resource.dediprogflasher import DediprogFlasher, NetworkDediprogFlasher
from labgrid.driver.dediprogflashdriver import DediprogFlashDriver


def test_dediprog_network_create(target):
    r = NetworkDediprogFlasher(target, name=None, vcc="3.5V", host="localhost")
    assert isinstance(r, NetworkDediprogFlasher)
    d = DediprogFlashDriver(target, name=None)
    assert isinstance(d, DediprogFlashDriver)


def test_dediprog_driver_create(target):
    r = DediprogFlasher(target, name=None, vcc="2.5V")
    assert isinstance(r, DediprogFlasher)

    d = DediprogFlashDriver(target, name=None)
    assert isinstance(d, DediprogFlashDriver)


def test_dediprog_driver_map_vcc(target):
    r = DediprogFlasher(target, name=None, vcc="2.5V")
    d = DediprogFlashDriver(target, name=None)

    m = d.map_vcc()
    assert isinstance(m, str)
    assert m == "1"


def test_dediprog_driver_map_vcc_inv(target):
    with pytest.raises(ValueError):
        r = DediprogFlasher(target, name=None, vcc="2.6V")



from labgrid.binding import BindingState
from labgrid.driver import BareboxDriver, UBootDriver, ShellDriver
from labgrid.driver.fake import FakeConsoleDriver, FakePowerDriver
from labgrid.strategy import Strategy, BareboxStrategy, UBootStrategy


def test_create_barebox(target):
    console = FakeConsoleDriver(target, "console")
    power = FakePowerDriver(target, "power")
    barebox = BareboxDriver(target, "barebox")
    shell = ShellDriver(target, "shell", prompt="root@dummy", login_prompt="login:", username="root")
    s = BareboxStrategy(target, "strategy")

    assert isinstance(s, Strategy)
    assert target.get_driver(BareboxStrategy) is s
    assert target.get_driver(Strategy) is s

    assert s.state is BindingState.bound


def test_create_uboot(target):
    console = FakeConsoleDriver(target, "console")
    power = FakePowerDriver(target, "power")
    barebox = UBootDriver(target, "uboot")
    shell = ShellDriver(target, "shell", prompt="root@dummy", login_prompt="login:", username="root")
    s = UBootStrategy(target, "strategy")

    assert isinstance(s, Strategy)
    assert target.get_driver(UBootStrategy) is s
    assert target.get_driver(Strategy) is s

    assert s.state is BindingState.bound



import abc


class FileTransferProtocol(abc.ABC):
    @abc.abstractmethod
    def put(self, filename: str, remotepath: str):
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, filename: str, destination: str):
        raise NotImplementedError



import pytest

from labgrid import Target
from labgrid.driver import BareboxDriver
from labgrid.driver.fake import FakeConsoleDriver
from labgrid.protocol import CommandProtocol, LinuxBootProtocol
from labgrid.driver import ExecutionError


class TestBareboxDriver:
    def test_create(self):
        t = Target('dummy')
        cp = FakeConsoleDriver(t, "console")
        d = BareboxDriver(t, "barebox")
        assert (isinstance(d, BareboxDriver))
        assert (isinstance(d, CommandProtocol))
        assert (isinstance(d, LinuxBootProtocol))

    def test_barebox_run(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = BareboxDriver(t, "barebox")
        d = t.get_driver(BareboxDriver, activate=False)
        # mock for d._run('echo $global.loglevel')
        d._run = mocker.MagicMock(return_value=(['7'], [], 0))
        t.activate(d)
        d._run = mocker.MagicMock(return_value=(['success'], [], 0))
        res = d.run_check("test")
        assert res == ['success']
        res = d.run("test")
        assert res == (['success'], [], 0)

    def test_barebox_run_error(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = BareboxDriver(t, "barebox")
        d = t.get_driver(BareboxDriver, activate=False)
        # mock for d._run('echo $global.loglevel')
        d._run = mocker.MagicMock(return_value=(['7'], [], 0))
        t.activate(d)
        d._run = mocker.MagicMock(return_value=(['error'], [], 1))
        with pytest.raises(ExecutionError):
            res = d.run_check("test")
        res = d.run("test")
        assert res == (['error'], [], 1)

    def test_barebox_boot(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = BareboxDriver(t, "barebox", boot_command='boot -v foo')
        d = t.get_driver(BareboxDriver, activate=False)
        # mock for d._run('echo $global.loglevel')
        d._run = mocker.MagicMock(return_value=(['7'], [], 0))
        t.activate(d)
        d._run = mocker.MagicMock(return_value=(['success'], [], 0))
        d.boot()
        assert d.console.txq.pop() == b"boot -v foo\n"
        d.boot(name='bar')
        assert d.console.txq.pop() == b"boot -v bar\n"



import pytest
import subprocess
import textwrap

from labgrid.util.helper import processwrapper


def test_processwrapper_passes():
    processwrapper.check_output(["true"])


def test_processwrapper_fails():
    with pytest.raises(subprocess.CalledProcessError):
        processwrapper.check_output(["false"])


def test_processwrapper_output(tmpdir):
    content = textwrap.dedent(
        """\
        A
        B
        C
        """
    )

    tmpfile = tmpdir.join("output.txt")
    tmpfile.write(content)

    output = processwrapper.check_output(["cat", str(tmpfile)])

    assert output.decode("utf-8") == content


def test_processwrapper_input():
    content = textwrap.dedent(
        """\
        A
        B
        C
        """
    )

    output = processwrapper.check_output(["cat", "-"], input=content.encode("utf-8"))

    assert output.decode("utf-8") == content


def test_processwrapper_empty_input():
    output = processwrapper.check_output(["cat", "-"], input=b"")
    assert output.decode("utf-8") == ""



import logging
from pprint import pprint

from labgrid import Environment
from labgrid.logging import basicConfig, StepLogger


# enable debug logging
basicConfig(level=logging.DEBUG)

# show labgrid steps on the console
StepLogger.start()


e = Environment("nm.env")
t = e.get_target()
d = t.get_driver("NetworkInterfaceDriver")

# based on https://developer.gnome.org/NetworkManager/stable/ch01.html, but adapted to python dicts
s_client = {
    "connection": {
        "type": "802-11-wireless",
    },
    "802-11-wireless": {
        "mode": "infrastructure",
        "ssid": "local-rpi",
    },
    "802-11-wireless-security": {
        "key-mgmt": "wpa-psk",
        "psk": "obMinwyurArc5",
    },
    "ipv4": {
        "method": "auto",
        "ignore-auto-dns": True,
        "ignore-auto-routes": True,
        "never-default": True,
    },
    "ipv6": {
        "method": "link-local",
    },
}
s_ap = {
    "connection": {
        "type": "802-11-wireless",
    },
    "802-11-wireless": {
        "mode": "ap",
        "ssid": "local-rpi",
    },
    "802-11-wireless-security": {
        "key-mgmt": "wpa-psk",
        "psk": "obMinwyurArc5",
    },
    "ipv4": {
        #'method': "auto",
        #'method': "link-local",
        "method": "shared",
        "addresses": ["172.16.0.2/29"],
    },
    "ipv6": {
        "method": "link-local",
    },
}

d.disable()
d.wait_state("disconnected")
print("access points after scan")
pprint(d.get_access_points())

d.configure(s_ap)
d.wait_state("activated")
print("settings in AP mode")
pprint(d.get_settings())
print("state in AP mode")
pprint(d.get_state())

# d.configure(s_client)
# d.wait_state('activated')
# print("settings in client mode")
# pprint(d.get_settings())
# print("state in client mode")
# pprint(d.get_state())



import abc
import re

import attr
import pytest

from labgrid import Target, target_factory
from labgrid.binding import BindingError
from labgrid.resource import Resource
from labgrid.driver import Driver
from labgrid.strategy import Strategy
from labgrid.exceptions import NoSupplierFoundError, NoDriverFoundError, NoResourceFoundError, NoStrategyFoundError


# test basic construction
def test_instanziation():
    t = Target("name")
    assert (isinstance(t, Target))


def test_get_resource(target):
    class A(Resource):
        pass

    a = A(target, "aresource")
    assert isinstance(target.get_resource(A), A)
    assert target.get_resource(A) is a
    assert target.get_resource(A, name="aresource") is a

    # make sure resources named "default" are prioritized
    b = A(target, "default")
    assert target.get_resource(A) is b
    assert target.get_resource(A, name="aresource") is a

def test_get_resource_multiple_no_default(target):
    class A(Resource):
        pass

    a = A(target, "aresource")
    b = A(target, "default")
    with pytest.raises(NoResourceFoundError) as excinfo:
        target.get_resource(A, name="nosuchresource")

def test_get_resource_multiple_with_default(target):
    class A(Resource):
        pass

    class B(Resource):
        pass

    a = A(target, "aresource")
    adef = A(target, "default")
    b = B(target, "bresource")
    bdef = B(target, "default")

    assert target.get_resource(A) is adef
    assert target.get_resource(B) is bdef
    assert target.get_resource(A, name="aresource") is a
    assert target.get_resource(B, name="bresource") is b

def test_get_driver(target):
    class A(Driver):
        pass

    a = A(target, "adriver")
    assert isinstance(target.get_driver(A), A)
    assert target.get_driver(A) is a
    assert target.get_driver(A, name="adriver") is a


def test_getitem(target):
    class AProtocol(abc.ABC):
        pass

    class A(Driver, AProtocol):
        pass

    class B(Driver):
        pass

    a = A(target, "adriver")
    target.activate(a)
    assert isinstance(target[A], A)
    assert target[A] is a
    assert target[AProtocol] is a
    assert target[A, "adriver"] is a
    assert target[AProtocol, "adriver"] is a
    with pytest.raises(NoDriverFoundError) as excinfo:
        target[A, "bdriver"]
    assert "matching resources with other names" in excinfo.value.msg
    with pytest.raises(NoDriverFoundError) as excinfo:
        target[B, "adriver"]
    assert re.match(f"no active .*? driver named '{a.name}' found in Target",
                    excinfo.value.msg)

    a2 = A(target, None)
    target.activate(a2)
    with pytest.raises(NoDriverFoundError) as excinfo:
        target[A]
    assert "multiple active drivers" in excinfo.value.msg


def test_no_resource(target):
    with pytest.raises(NoResourceFoundError):
        target.get_resource(Target)


def test_no_driver(target):
    with pytest.raises(NoDriverFoundError):
        target.get_driver(Target)


# test alternative suppliers
class ResourceA(Resource):
    pass


class ResourceB(Resource):
    pass


class DriverWithA(Driver):
    bindings = {"res": ResourceA}


class DriverWithASet(Driver):
    bindings = {"res": {ResourceA}, }


class DriverWithAB(Driver):
    bindings = {"res": {ResourceA, ResourceB}, }


def test_suppliers_a(target):
    ra = ResourceA(target, "resource")
    d = DriverWithA(target, "resource")
    assert d.res is ra


def test_suppliers_aset(target):
    ra = ResourceA(target, "resource")
    d = DriverWithASet(target, "driver")
    assert d.res is ra


def test_suppliers_ab_a(target):
    ra = ResourceA(target, "resource")
    d = DriverWithAB(target, "driver")
    assert d.res is ra


def test_suppliers_ab_b(target):
    rb = ResourceB(target, "resource")
    d = DriverWithAB(target, "driver")
    assert d.res is rb


def test_suppliers_ab_both(target):
    ra = ResourceA(target, "resource_a")
    rb = ResourceB(target, "resource_b")
    with pytest.raises(NoSupplierFoundError):
        d = DriverWithAB(target, "driver")


def test_suppliers_ab_missing(target):
    with pytest.raises(NoSupplierFoundError):
        d = DriverWithAB(target, "driver")


def test_suppliers_unexpected_binding(target):
    ra = ResourceA(target, "resource")
    target.set_binding_map({"res": "resource", "unexpected": "foo"})
    with pytest.raises(BindingError) as excinfo:
        DriverWithA(target, "driver")
    assert "got unexpected bindings" in excinfo.value.msg


class DriverWithNamedA(Driver):
    bindings = {
        "res": Driver.NamedBinding(ResourceA),
    }


def test_suppliers_named_a(target):
    ra = ResourceA(target, "resource")
    target.set_binding_map({"res": "resource"})
    d = DriverWithNamedA(target, "driver")
    assert d.res is ra


class DriverWithMultiA(Driver):
    bindings = {
        "res1": ResourceA,
        "res2": ResourceA,
    }


def test_suppliers_multi_a(target):
    ra1 = ResourceA(target, "resource1")
    with pytest.raises(BindingError) as excinfo:
        DriverWithMultiA(target, "driver")
    assert "duplicate bindings" in excinfo.value.msg


def test_suppliers_multi_a_explict(target):
    ra1 = ResourceA(target, "resource1")
    ra2 = ResourceA(target, "resource2")
    target.set_binding_map({
        "res1": "resource1",
        "res2": "resource2",
    })
    d = DriverWithMultiA(target, "driver")
    assert d.res1 is ra1
    assert d.res2 is ra2


class DriverWithNamedMultiA(Driver):
    bindings = {
        "res1": Driver.NamedBinding(ResourceA),
        "res2": Driver.NamedBinding(ResourceA),
    }


def test_suppliers_multi_named_a(target):
    ra1 = ResourceA(target, "resource1")
    ra2 = ResourceA(target, "resource2")
    target.set_binding_map({
        "res1": "resource1",
        "res2": "resource2",
    })
    d = DriverWithNamedMultiA(target, "driver")
    assert d.res1 is ra1
    assert d.res2 is ra2


# test optional bindings

class DriverWithOptionalA(Driver):
    bindings = {"res": {ResourceA, None}, }


class DriverWithOptionalAB(Driver):
    bindings = {"res": {ResourceA, ResourceB, None}, }


def test_suppliers_optional_a(target):
    ra = ResourceA(target, "resource")
    d = DriverWithOptionalA(target, "driver")
    assert d.res is ra


def test_suppliers_optional_a_missing(target):
    rb = ResourceB(target, "resource")
    d = DriverWithOptionalA(target, "driver")
    assert d.res is None


def test_suppliers_optional_ab_a(target):
    ra = ResourceA(target, "resource")
    d = DriverWithOptionalAB(target, "driver")
    assert d.res is ra


class DriverWithOptionalNamedA(Driver):
    bindings = {
        "res": Driver.NamedBinding({ResourceA, None}),
    }


def test_suppliers_optional_named_a(target):
    ra = ResourceA(target, "resource")
    target.set_binding_map({"res": "resource"})
    d = DriverWithOptionalNamedA(target, "driver")
    assert d.res is ra


def test_suppliers_optional_named_a_missing(target):
    rb = ResourceB(target, "resource")
    target.set_binding_map({"res": "resource"})
    d = DriverWithOptionalNamedA(target, "driver")
    assert d.res is None



class StrategyA(Strategy):
    bindings = {
        "drv": DriverWithA,
    }


def test_get_strategy(target):
    ra = ResourceA(target, "resource")
    d = DriverWithA(target, "driver")

    with pytest.raises(NoStrategyFoundError):
        target.get_strategy()

    s1 = StrategyA(target, "s1")
    assert target.get_strategy() is s1

    s2 = StrategyA(target, "s2")
    with pytest.raises(NoStrategyFoundError):
        target.get_strategy()


# test nested resource creation
@attr.s(eq=False)
class DiscoveryResource(Resource):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        ResourceA(self.target, "resource")

def test_nested(target):
    rd = DiscoveryResource(target, "discovery")
    d = DriverWithAB(target, "driver")
    assert isinstance(d.res, ResourceA)

# Test retrieving drivers, resources and protocols by name

def test_get_by_string(target):
    class AProtocol(abc.ABC):
        pass

    @target_factory.reg_driver
    class A(Driver, AProtocol):
        pass

    @target_factory.reg_driver
    class C(Resource):
        pass

    a = A(target, None)
    target.activate(a)
    assert target.get_driver('A') == a
    assert target.get_active_driver('A') == a

    c = C(target, None)
    assert target.get_resource('C') == c

    assert target['AProtocol'] == a

    with pytest.raises(KeyError):
        target.get_driver("nosuchdriver")

# Test priorities

def test_get_by_diff_priority(target):
    class AProtocol(abc.ABC):
        pass

    @attr.s
    class A(Driver, AProtocol):
        priorities = {AProtocol: -10}

    @attr.s
    class C(Driver, AProtocol):
        priorities = {AProtocol: 10}
        pass

    a = A(target, None)
    c = C(target, None)
    target.activate(a)
    target.activate(c)

    assert target.get_driver(AProtocol) == c

def test_get_by_same_priority(target):
    class AProtocol(abc.ABC):
        pass

    @attr.s
    class A(Driver, AProtocol):
        priorities = {AProtocol: 10}

    @attr.s
    class C(Driver, AProtocol):
        priorities = {AProtocol: 10}
        pass

    a = A(target, None)
    c = C(target, None)
    target.activate(a)
    target.activate(c)

    with pytest.raises(NoDriverFoundError) as e_info:
        target.get_driver(AProtocol)
    assert "multiple drivers matching" in str(e_info.value)

def test_get_by_default_priority(target):
    class AProtocol(abc.ABC):
        pass

    @attr.s
    class A(Driver, AProtocol):
        pass

    @attr.s
    class C(Driver, AProtocol):
        pass

    a = A(target, None)
    c = C(target, None)
    target.activate(a)
    target.activate(c)

    with pytest.raises(NoDriverFoundError) as e_info:
        target.get_driver(AProtocol)
    assert "multiple drivers matching" in str(e_info.value)

def test_target_deactivate_by_string(target):

    @target_factory.reg_driver
    @attr.s
    class ADea(Driver):
        pass

    a = ADea(target, None)
    target.activate(a)
    target.deactivate("ADea")

    assert a == target.get_driver(ADea, activate=False)

def test_target_activate_by_string(target):

    @target_factory.reg_driver
    @attr.s
    class AActiv(Driver):
        pass

    a = AActiv(target, None)
    target.activate("AActiv")

    assert a == target.get_active_driver(AActiv)

def test_allow_binding_by_different_protocols(target):
    class ADiffProtocol(abc.ABC):
        pass

    class BDiffProtocol(abc.ABC):
        pass

    class DiffDriver(Driver, ADiffProtocol, BDiffProtocol):
        pass

    class DiffStrategy(Strategy):
        bindings = {
            "a": ADiffProtocol,
            "b": BDiffProtocol
        }

    d = DiffDriver(target, None)
    s = DiffStrategy(target, None)

    assert s.a == d
    assert s.b == d

def test_allow_optional_binding_by_different_protocols(target):
    class AOpt1DiffProtocol(abc.ABC):
        pass

    class BOpt1DiffProtocol(abc.ABC):
        pass

    class Opt1DiffDriver(Driver, AOpt1DiffProtocol):
        pass

    class Opt1DiffStrategy(Strategy):
        bindings = {
            "a": AOpt1DiffProtocol,
            "b": {BOpt1DiffProtocol, None}
        }

    d = Opt1DiffDriver(target, None)
    s = Opt1DiffStrategy(target, None)

    assert s.a == d
    assert s.b == None

def test_allow_optional_available_binding_by_different_protocols(target):
    class AOpt2DiffProtocol(abc.ABC):
        pass

    class BOpt2DiffProtocol(abc.ABC):
        pass

    class Opt2DiffDriver(Driver, AOpt2DiffProtocol, BOpt2DiffProtocol):
        pass

    class Opt2DiffStrategy(Strategy):
        bindings = {
            "a": {AOpt2DiffProtocol, None},
            "b": {BOpt2DiffProtocol, None}
        }

    d = Opt2DiffDriver(target, None)
    s = Opt2DiffStrategy(target, None)

    assert s.a == d
    assert s.b == d

def test_allow_optional_no_available_binding_by_different_protocols(target):
    class AOpt3DiffProtocol(abc.ABC):
        pass

    class BOpt3DiffProtocol(abc.ABC):
        pass

    class Opt3DiffDriver(Driver):
        pass

    class Opt3DiffStrategy(Strategy):
        bindings = {
            "a": {AOpt3DiffProtocol, None},
            "b": {BOpt3DiffProtocol, None}
        }

    d = Opt3DiffDriver(target, None)
    s = Opt3DiffStrategy(target, None)

    assert s.a == None
    assert s.b == None

def test_allow_optional_no_double_same_protocol_by_different_protocols(target):
    class AOpt4DiffProtocol(abc.ABC):
        pass

    class Opt4DiffDriver(Driver, AOpt4DiffProtocol):
        pass

    class Opt4DiffStrategy(Strategy):
        bindings = {
            "a": {AOpt4DiffProtocol, None},
            "b": {AOpt4DiffProtocol, None},
            "c": {AOpt4DiffProtocol, None},
            "d": {AOpt4DiffProtocol, None},
        }

    d1 = Opt4DiffDriver(target, name="driver1")
    d2 = Opt4DiffDriver(target, name="driver2")
    target.set_binding_map({"c": "driver1", "d": "driver2"})
    s = Opt4DiffStrategy(target, None)

    assert s.a is None
    assert s.b is None
    assert s.c == d1
    assert s.d == d2

def test_get_bound_resources(target):
    class AResource(Resource):
        pass

    class ADriver(Driver):
        bindings = {
            "a": AResource,
        }

    class BDriver(Driver):
        bindings = {
            "a": ADriver,
        }

    aresource = AResource(target, "aresource")
    adriver = ADriver(target, "adriver")
    bdriver = BDriver(target, "bdriver")

    assert bdriver.get_bound_resources() == {aresource}
    assert adriver.get_bound_resources() == {aresource}

    assert target.get_driver(ADriver, resource=aresource) 

def test_get_bound_multiple_resources(target):
    class AResource(Resource):
        pass

    class BResource(Resource):
        pass

    class ADriver(Driver):
        bindings = {
            "a": AResource,
        }

    class BDriver(Driver):
        bindings = {
            "b": BResource,
        }

    class CDriver(Driver):
        bindings = {
            "a": ADriver,
            "b": BDriver,
        }

    aresource = AResource(target, "aresource")
    bresource = BResource(target, "bresource")
    adriver = ADriver(target, "adriver")
    bdriver = BDriver(target, "bdriver")
    cdriver = CDriver(target, "bdriver")

    assert adriver.get_bound_resources() == {aresource}
    assert bdriver.get_bound_resources() == {bresource}
    assert cdriver.get_bound_resources() == {aresource, bresource}

    assert target.get_driver(ADriver, resource=aresource) 
    assert target.get_driver(BDriver, resource=bresource) 
    assert target.get_driver(CDriver, resource=aresource) 

    assert target.get_active_driver(ADriver, resource=aresource) 
    assert target.get_active_driver(BDriver, resource=bresource) 
    assert target.get_active_driver(CDriver, resource=aresource) 



import os

import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..step import step
from .common import Driver


@target_factory.reg_driver
@attr.s(eq=False)
class FileDigitalOutputDriver(Driver, DigitalOutputProtocol):
    """
    Two arbitrary string values false_repr and true_repr
    are defined as representations for False and True.
    These values are written to a file and read from it.
    If the file's content does not match any of the
    representations it defaults to False.
    A prime example for using this driver is Linux's sysfs.
    """

    filepath = attr.ib(validator=attr.validators.instance_of(str))
    false_repr = attr.ib(default="0\n", validator=attr.validators.instance_of(str))
    true_repr = attr.ib(default="1\n", validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        if not os.path.isfile(self.filepath):
            raise Exception(f"{self.filepath} is not a file.")

    @Driver.check_active
    @step()
    def get(self):
        with open(self.filepath) as fdes:
            from_file = fdes.read()
        return from_file == self.true_repr

    @Driver.check_active
    @step()
    def set(self, status):
        out = self.true_repr if status else self.false_repr
        with open(self.filepath, "w") as fdes:
            fdes.write(out)



import enum

import attr

from ..factory import target_factory
from ..step import step
from .common import Strategy, StrategyError, never_retry


class Status(enum.Enum):
    unknown = 0
    off = 1
    shell = 2


@target_factory.reg_driver
@attr.s(eq=False)
class ShellStrategy(Strategy):
    """ShellStrategy - Strategy to switch to shell"""
    bindings = {
        "power": "PowerProtocol",
        "console": "ConsoleProtocol",
        "shell": "ShellDriver",
    }

    status = attr.ib(default=Status.unknown)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @never_retry
    @step(args=['status'])
    def transition(self, status, *, step):  # pylint: disable=redefined-outer-name
        if not isinstance(status, Status):
            status = Status[status]
        if status == Status.unknown:
            raise StrategyError(f"can not transition to {status}")
        elif status == self.status:
            step.skip("nothing to do")
            return  # nothing to do
        elif status == Status.off:
            self.target.deactivate(self.console)
            self.target.activate(self.power)
            self.power.off()
        elif status == Status.shell:
            self.transition(Status.off)
            self.target.activate(self.console)
            self.power.cycle()
            self.target.activate(self.shell)
            self.shell.run("systemctl is-system-running --wait")
        else:
            raise StrategyError(
                f"no transition found from {self.status} to {status}"
            )
        self.status = status

    @never_retry
    @step(args=['status'])
    def force(self, status, *, step):  # pylint: disable=redefined-outer-name
        if not isinstance(status, Status):
            status = Status[status]
        if status == Status.unknown:
            raise StrategyError(f"can not force state {status}")
        elif status == Status.off:
            self.target.deactivate(self.shell)
            self.target.activate(self.power)
        elif status == Status.shell:
            self.target.activate(self.power)
            self.target.activate(self.shell)
        else:
            raise StrategyError(f"not setup found for {status}")
        self.status = status



import pytest
import warnings

from labgrid import Environment
from labgrid.exceptions import NoConfigFoundError, InvalidConfigError
from labgrid.protocol import ConsoleProtocol
from labgrid.resource import RawSerialPort


class TestEnvironment:
    def test_noconfig_instance(self):
        with pytest.raises(NoConfigFoundError):
            e = Environment()

    def test_instance(self, tmpdir):
        p = tmpdir.join("config.yaml")
        p.write(
            """
        targets:
          test1:
            drivers: {}
          test2:
            role: foo
            resources: {}
        """
        )
        e = Environment(str(p))
        assert (isinstance(e, Environment))

    def test_get_target(self, tmpdir):
        p = tmpdir.join("config.yaml")
        p.write(
            """
        targets:
          test1:
            drivers: {}
          test2:
            role: foo
            resources: {}
        """
        )
        e = Environment(str(p))
        assert (e.get_target("test1"))
        assert (e.get_target("test2"))

    def test_instance_invalid_yaml(self, tmpdir):
        p = tmpdir.join("config.yaml")
        p.write(
            """
        I a(m) no yaml:
          - keks
          cookie
        """
        )
        with pytest.raises(InvalidConfigError):
            e = Environment(str(p))

    def test_env_imports_yaml(self, tmpdir):
        import sys
        i = tmpdir.join("myimport.py")
        i.write(
"""
class Test:
    pass
"""
        )
        p = tmpdir.join("config.yaml")
        p.write(
    f"""
targets:
  main:
    drivers: {{}}
imports:
  - {str(i)}
"""
        )
        e = Environment(str(p))
        assert (isinstance(e, Environment))
        assert "myimport" in sys.modules
        import myimport
        t = myimport.Test()
        assert (isinstance(t, myimport.Test))

    def test_create_named_resources(self, tmpdir):
        p = tmpdir.join("config.yaml")
        p.write(
            """
        targets:
          test1:
            resources:
            - AndroidUSBFastboot:
                name: "fastboot"
                match: {}
            - RawSerialPort:
                port: "/dev/ttyUSB0"
                speed: 115200
        """
        )
        e = Environment(str(p))
        t = e.get_target("test1")

    def test_create_named_drivers(self, tmpdir):
        p = tmpdir.join("config.yaml")
        p.write(
            """
        targets:
          test1:
            resources:
            - AndroidUSBFastboot:
                name: "fastboot"
                match: {}
            - RawSerialPort:
                name: "serial_a"
                port: "/dev/ttyUSB0"
                speed: 115200
            - cls: RawSerialPort
              name: "serial_b"
              port: "/dev/ttyUSB0"
              speed: 115200
            drivers:
            - FakeConsoleDriver:
                name: "serial_a"
            - FakeConsoleDriver:
                name: "serial_b"
        """
        )
        e = Environment(str(p))
        t = e.get_target("test1")

    def test_create_multi_drivers(self, tmpdir):
        p = tmpdir.join("config.yaml")
        p.write(
            """
        targets:
          test1:
            resources:
            - RawSerialPort:
                name: "serial_a"
                port: "/dev/ttyUSB0"
                speed: 115200
            - RawSerialPort:
                name: "serial_b"
                port: "/dev/ttyUSB0"
                speed: 115200
            drivers:
            - SerialDriver:
                name: "serial_a"
                bindings:
                  port: "serial_a"
            - SerialDriver:
                name: "serial_b"
                bindings:
                  port: "serial_b"
        """
        )
        e = Environment(str(p))
        t = e.get_target("test1")
        r_a = t.get_resource(RawSerialPort, name="serial_a")
        r_b = t.get_resource(RawSerialPort, name="serial_b")
        assert r_a is not r_b
        d_a = t.get_driver(ConsoleProtocol, name="serial_a", activate=False)
        d_b = t.get_driver(ConsoleProtocol, name="serial_b", activate=False)
        assert d_a is not d_b

        assert d_a.port is r_a
        assert d_b.port is r_b

    def test_usbserialport_warning(self, tmpdir):
        p = tmpdir.join("config.yaml")
        p.write(
            """
        targets:
          test1:
            resources:
            - USBSerialPort:
                port: /dev/ttyS0
            drivers:
            - SerialDriver: {}
        """
        )
        e = Environment(str(p))
        with pytest.warns(UserWarning):
            t = e.get_target("test1")

    def test_usbserialport_no_warning(self, tmpdir):
        p = tmpdir.join("config.yaml")
        p.write(
            """
        targets:
          test1:
            resources:
            - USBSerialPort: {}
            drivers:
            - SerialDriver: {}
        """
        )
        e = Environment(str(p))
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            t = e.get_target("test1")



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class PyVISADevice(Resource):
    """The PyVISADevice describes a test stimuli device controlled  with PyVISA

    Args:
        type (str): device resource type following the pyVISA resource syntax, e.g. ASRL, TCPIP...
        url (str): device identifier on selected resource, e.g. <ip> for TCPIP resource
        backend (str, default=''): Visa library backend, e.g. '@sim' for pyvisa-sim backend
    """

    type = attr.ib(validator=attr.validators.instance_of(str))
    url = attr.ib(validator=attr.validators.instance_of(str))
    backend = attr.ib(default="", validator=attr.validators.instance_of(str))



"""
Class for connecting to a docker daemon running on the host machine.
"""

from enum import Enum

import attr

from labgrid.factory import target_factory
from labgrid.driver.common import Driver
from labgrid.resource.docker import DockerConstants
from labgrid.protocol.powerprotocol import PowerProtocol


class PullPolicy(Enum):
    """Pull policy for the DockerDriver.

    Modelled after ``podman run --pull`` / ``docker run --pull``.

    * always: Always pull the image and throw an error if the pull fails.
    * missing: Pull the image only when the image is not in the local
      containers storage. Throw an error if no image is found and the pull
      fails.
    * never: Never pull the image but use the one from the local containers
      storage. Throw an error if no image is found.
    * newer: **Note** not supported by the driver, and therefore not
      implemented.
    """

    Always = "always"
    Missing = "missing"
    Never = "never"


def pull_policy_converter(value):
    if isinstance(value, PullPolicy):
        return value
    try:
        return PullPolicy(value)
    except ValueError:
        raise ValueError(f"Invalid pull policy: {value}")


@target_factory.reg_driver
@attr.s(eq=False)
class DockerDriver(PowerProtocol, Driver):
    """The DockerDriver is used to create docker containers.
    This is done via communication with a docker daemon.

    When a container is created the container is labeled with an
    cleanup strategy identifier. Currently only one strategy is
    implemented.  This strategy simply deletes all labgrid created
    containers before each test run. This is to ensure cleanup of
    dangling containers from crashed tests or hanging containers.

    Image pruning is not done by the driver.

    For detailed information about the arguments see the
    "Docker SDK for Python" documentation
    https://docker-py.readthedocs.io/en/stable/containers.html#container-objects

    Args:
        bindings (dict): The labgrid bindings
    Args passed to docker.create_container:
        image_uri (str): The uri of the image to fetch
        pull (str): Pull policy. Default policy is ``always`` for backward
        compatibility concerns
        command (str): The command to execute once container has been created
        volumes (list): The volumes to declare
        environment (list): Docker environment variables to set
        host_config (dict): Docker host configuration parameters
        network_services (list): Sequence of dicts each specifying a network \
                                 service that the docker container exposes.

    """

    bindings = {"docker_daemon": {"DockerDaemon"}}
    image_uri = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    pull = attr.ib(default=PullPolicy.Always, converter=pull_policy_converter)
    command = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    volumes = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(list)))
    container_name = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    environment = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(list)))
    host_config = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(dict)))
    network_services = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(list)))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._client = None
        self._container = None

    def on_activate(self):
        """On activation:
        1. Import docker module (_client and _container remain available)
        2. Connect to the docker daemon
        3. Pull requested image from docker registry if needed
        4. Create the new container according to parameters from conf
        """
        import docker

        self._client = docker.DockerClient(base_url=self.docker_daemon.docker_daemon_url)

        if self.pull == PullPolicy.Always:
            self._client.images.pull(self.image_uri)
        elif self.pull == PullPolicy.Missing:
            try:
                self._client.images.get(self.image_uri)
            except docker.errors.ImageNotFound:
                self._client.images.pull(self.image_uri)
        elif self.pull == PullPolicy.Never:
            self._client.images.get(self.image_uri)

        self._container = self._client.api.create_container(
            self.image_uri,
            command=self.command,
            volumes=self.volumes,
            name=self.container_name,
            environment=self.environment,
            labels={DockerConstants.DOCKER_LG_CLEANUP_LABEL: DockerConstants.DOCKER_LG_CLEANUP_TYPE_AUTO},
            host_config=self._client.api.create_host_config(**self.host_config),
        )

    def on_deactivate(self):
        """Remove container after use"""
        self._client.api.remove_container(self._container.get("Id"), force=True)
        self._client = None
        self._container = None

    def on(self):
        """Start the container created during activation"""
        self._client.api.start(container=self._container.get("Id"))

    def off(self):
        """Stop the container created during activation"""
        self._client.api.stop(container=self._container.get("Id"))

    def cycle(self):
        """Cycle the docker container by stopping and starting it"""
        self.off()
        self.on()



import pytest

import grpc
import labgrid.remote.generated.labgrid_coordinator_pb2_grpc as labgrid_coordinator_pb2_grpc
import labgrid.remote.generated.labgrid_coordinator_pb2 as labgrid_coordinator_pb2


@pytest.fixture(scope="function")
def channel_stub():
    import queue

    queue = queue.Queue()

    channel = grpc.insecure_channel("127.0.0.1:20408")
    stub = labgrid_coordinator_pb2_grpc.CoordinatorStub(channel)

    def generate_startup(queue):
        msg = labgrid_coordinator_pb2.ClientInMessage()
        msg.startup.version = "2.0.0"
        msg.startup.name = "testclient"
        messages = [msg]
        for msg in messages:
            yield msg
        while True:
            msg = queue.get()
            yield msg
            queue.task_done()

    stream = stub.ClientStream(generate_startup(queue))
    yield stub
    channel.close()


@pytest.fixture(scope="function")
def coordinator_place(channel_stub):
    name = "test"
    place = labgrid_coordinator_pb2.AddPlaceRequest(name=name)
    res = channel_stub.AddPlace(place)
    assert res, f"There was an error: {res}"
    return channel_stub


def test_startup(coordinator):
    pass


def test_coordinator_add_place(coordinator, channel_stub):
    name = "test"
    place = labgrid_coordinator_pb2.AddPlaceRequest(name=name)
    res = channel_stub.AddPlace(place)
    assert res, f"There was an error: {res}"


def test_coordinator_del_place(coordinator, channel_stub):
    name = "test"
    place = labgrid_coordinator_pb2.AddPlaceRequest(name=name)
    res = channel_stub.AddPlace(place)
    assert res, f"There was an error: {res}"
    place = labgrid_coordinator_pb2.DeletePlaceRequest(name=name)
    res = channel_stub.DeletePlace(place)
    assert res, f"There was an error: {res}"


def test_coordinator_get_places(coordinator, channel_stub):
    name = "test"
    place = labgrid_coordinator_pb2.AddPlaceRequest(name=name)
    res = channel_stub.AddPlace(place)
    assert res, f"There was an error: {res}"
    name = "test2"
    place = labgrid_coordinator_pb2.AddPlaceRequest(name=name)
    res = channel_stub.AddPlace(place)
    assert res, f"There was an error: {res}"

    request = labgrid_coordinator_pb2.GetPlacesRequest()
    res = channel_stub.GetPlaces(request)

    from labgrid.remote.common import Place

    places = set()
    names = set()
    for pb2 in res.places:
        place = Place.from_pb2(pb2)
        places.add(place)
        names.add(place.name)

    assert len(places) == 2, f"Returned places not two: {places}"
    assert set(names) == {"test", "test2"}, f"There was an error: {res}"


def test_coordinator_exporter_session(coordinator, channel_stub):
    import queue

    queue = queue.Queue()

    def generate_startup(queue):
        msg = labgrid_coordinator_pb2.ExporterInMessage()
        msg.startup.version = "2.0.0"
        msg.startup.name = "testporter"
        messages = [msg]
        for msg in messages:
            yield msg
        while True:
            msg = queue.get()
            yield msg
            queue.task_done()

    coordinator = channel_stub.ExporterStream(generate_startup(queue), wait_for_ready=True)


def test_coordinator_place_acquire(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.AcquirePlace(labgrid_coordinator_pb2.AcquirePlaceRequest(placename="test"))
    assert res


def test_coordinator_place_acquire_release(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.AcquirePlace(labgrid_coordinator_pb2.AcquirePlaceRequest(placename="test"))
    assert res
    res = stub.ReleasePlace(labgrid_coordinator_pb2.ReleasePlaceRequest(placename="test"))
    assert res


def test_coordinator_place_add_alias(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.AddPlaceAlias(labgrid_coordinator_pb2.AddPlaceAliasRequest(placename="test", alias="testalias"))
    assert res


def test_coordinator_place_add_remove_alias(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.AddPlaceAlias(labgrid_coordinator_pb2.AddPlaceAliasRequest(placename="test", alias="testalias"))
    assert res
    res = stub.DeletePlaceAlias(labgrid_coordinator_pb2.DeletePlaceAliasRequest(placename="test", alias="testalias"))
    assert res


def test_coordinator_place_set_tags(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.SetPlaceTags(labgrid_coordinator_pb2.SetPlaceTagsRequest(placename="test", tags={"one": "two"}))
    assert res


def test_coordinator_place_set_comment(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.SetPlaceComment(labgrid_coordinator_pb2.SetPlaceCommentRequest(placename="test", comment="testcomment"))
    assert res


def test_coordinator_place_add_match(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.AddPlaceMatch(
        labgrid_coordinator_pb2.AddPlaceMatchRequest(placename="test", pattern="this/test/pattern")
    )
    assert res


def test_coordinator_place_add_delete_match(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.AddPlaceMatch(
        labgrid_coordinator_pb2.AddPlaceMatchRequest(placename="test", pattern="this/test/pattern")
    )
    assert res
    res = stub.DeletePlaceMatch(
        labgrid_coordinator_pb2.DeletePlaceMatchRequest(placename="test", pattern="this/test/pattern")
    )
    assert res


def test_coordinator_place_allow(coordinator, coordinator_place):
    stub = coordinator_place
    res = stub.AcquirePlace(labgrid_coordinator_pb2.AcquirePlaceRequest(placename="test"))
    assert res
    res = stub.AllowPlace(labgrid_coordinator_pb2.AllowPlaceRequest(placename="test", user="othertest"))
    assert res


def test_coordinator_create_reservation(coordinator, coordinator_place):
    tags = {"board": "test"}
    stub = coordinator_place
    res = stub.SetPlaceTags(labgrid_coordinator_pb2.SetPlaceTagsRequest(placename="test", tags=tags))
    assert res
    res = stub.CreateReservation(
        labgrid_coordinator_pb2.CreateReservationRequest(
            filters={
                "main": labgrid_coordinator_pb2.Reservation.Filter(filter={"board": "test"}),
            },
            prio=1.0,
        )
    )
    assert res
    res: labgrid_coordinator_pb2.CreateReservationResponse
    assert len(res.reservation.token) > 0



import os
from typing import Optional
import attr

from .target import Target
from .config import Config


@attr.s(eq=False)
class Environment:
    """An environment encapsulates targets."""
    config_file = attr.ib(
        default="config.yaml", validator=attr.validators.instance_of(str)
    )
    interact = attr.ib(default=input, repr=False)

    def __attrs_post_init__(self):
        self.targets = {}

        self.config = Config(self.config_file)

        for user_import in self.config.get_imports():
            import importlib.util
            from importlib.machinery import SourceFileLoader
            import sys

            if user_import.endswith('.py'):
                module_name = os.path.basename(user_import)[:-3]
                loader = SourceFileLoader(module_name, user_import)
                spec = importlib.util.spec_from_loader(loader.name, loader)
                module = importlib.util.module_from_spec(spec)
                loader.exec_module(module)
            else:
                module_name = user_import
                module = importlib.import_module(user_import)
            sys.modules[module_name] = module

    def get_target(self, role: str = 'main') -> Optional[Target]:
        """Returns the specified target or None if not found.

        Each target is initialized as needed.
        """
        from . import target_factory

        if role not in self.targets:
            config = self.config.get_targets().get(role)
            if not config:
                return None
            target = target_factory.make_target(role, config, env=self)
            self.targets[role] = target

        return self.targets[role]

    def get_features(self):
        return self.config.get_features()

    def get_target_features(self):
        flags = set()
        for value in self.config.get_targets().values():
            flags = flags | set(value.get('features', {}))
        return flags

    def cleanup(self):
        for target in self.targets:
            self.targets[target].cleanup()



import enum
from functools import wraps
from typing import Any, Dict

import attr


@attr.s(eq=False)
class StateError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))


@attr.s(eq=False)
class BindingError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))


@enum.unique
class BindingState(enum.Enum):
    error = -1
    idle = 0
    bound = 1
    active = 2


@attr.s(eq=False)
class BindingMixin:
    """
    Handles the binding and activation of drivers and their supplying resources
    and drivers.

    One client can be bound to many suppliers, and one supplier can be bound by
    many clients.

    Conflicting access to one supplier can be avoided by deactivating
    conflicting clients before activation (using the resolve_conflicts
    callback).
    """

    bindings: Dict[str, Any] = {}

    # these are controlled by the Target
    target = attr.ib()
    name = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)))
    state = attr.ib(default=BindingState.idle, init=False)

    def __attrs_post_init__(self):
        self.suppliers = set()
        self.clients = set()
        target = self.target
        if target is not None:
            # bind will set it again if successful
            self.target = None
            target.bind(self)
            assert self.target is not None

    @property
    def display_name(self):
        if self.name:
            return f"{self.__class__.__name__}(target={self.target.name}, name={self.name})"

        return f"{self.__class__.__name__}(target={self.target.name})"

    def on_supplier_bound(self, supplier):
        """Called by the Target after a new supplier has been bound"""
        pass

    def on_client_bound(self, client):
        """Called by the Target after a new client has been bound"""
        pass

    def on_activate(self):
        """Called by the Target when this object has been activated"""
        pass

    def on_deactivate(self):
        """Called by the Target when this object has been deactivated"""
        pass

    def resolve_conflicts(self, client):
        """
        Called by the Target to allow this object to deactivate conflicting
        clients.
        """
        pass

    def get_bound_resources(self):
        """
        Called by the Target to find the correct driver corresponding to a resource
        """
        raise NotImplementedError("get_bound_resources not implemented!")

    @classmethod
    def check_active(cls, func):
        @wraps(func)
        def wrapper(self, *_args, **_kwargs):
            if self.state is not BindingState.active:
                raise StateError(
                    f'{self} has not been activated, {func.__qualname__} cannot be called in state "{self.state.name}"'  # pylint: disable=line-too-long
                )
            return func(self, *_args, **_kwargs)

        return wrapper

    @classmethod
    def check_bound(cls, func):
        @wraps(func)
        def wrapper(self, *_args, **_kwargs):
            if self.state is BindingState.active:
                raise StateError(f"{self} is active, but must be deactivated to call {func.__qualname__}")
            elif self.state is not BindingState.bound:
                raise StateError(
                    f'{self} has not been bound, {func.__qualname__} cannot be called in state "{self.state.name}"'  # pylint: disable=line-too-long
                )
            return func(self, *_args, **_kwargs)

        return wrapper

    class NamedBinding:
        """
        Marks a binding (or binding set) as requiring an explicit name.
        """

        def __init__(self, value):
            self.value = value

        def __repr__(self):
            return f"Binding.Named({repr(self.value)})"



import time
import pexpect

from ..util import PtxExpect, Timeout
from ..step import step
from .common import Driver


class ConsoleExpectMixin:
    """
    Console driver mixin to implement the read, write, expect and sendline methods. It uses
    the internal _read and _write methods.

    The class using the ConsoleExpectMixin must provide a logger and a txdelay attribute.
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._expect = PtxExpect(self)

    @Driver.check_active
    @step(result=True, tag='console')
    def read(self, size=1, timeout=0.0, max_size=None):
        res = self._read(size=size, timeout=timeout, max_size=max_size)
        if max_size:
            self.logger.debug("Read %i bytes: %s, timeout %.2f, requested size %i, max size %i",
                              len(res), res, timeout, size, max_size)
        else:
            self.logger.debug("Read %i bytes: %s, timeout %.2f, requested size %i",
                              len(res), res, timeout, size)
        return res

    @Driver.check_active
    @step(args=['data'], tag='console')
    def write(self, data):
        if self.txdelay:
            self.logger.debug("Write %i bytes: %s (with %fs txdelay per %i bytes)",
                              len(data), data, self.txdelay, self.txchunk)

            count = 0
            for i in range(0, len(data), self.txchunk):
                time.sleep(self.txdelay)
                count += self._write(data[i:i+self.txchunk])

            return count

        self.logger.debug("Write %i bytes: %s", len(data), data)
        return self._write(data)

    @Driver.check_active
    def sendline(self, line):
        self._expect.sendline(line)

    @Driver.check_active
    def sendcontrol(self, char):
        self._expect.sendcontrol(char)

    @Driver.check_active
    @step(args=['pattern'], result=True)
    def expect(self, pattern, timeout=-1):
        index = self._expect.expect(pattern, timeout=timeout)
        return index, self._expect.before, self._expect.match, self._expect.after

    @Driver.check_active
    @step(args=['quiet_time'])
    def settle(self, quiet_time, timeout=120.0) -> bool:
        t = Timeout(timeout)
        while not t.expired:
            try:
                self.read(timeout=quiet_time)
            except pexpect.TIMEOUT:
                return True
        return False

    def resolve_conflicts(self, client):
        for other in self.clients:
            if other is client:
                continue
            self.target.deactivate(other)



import pytest
import requests

from labgrid.resource import LXAIOBusPIO
from labgrid.driver import LXAIOBusPIODriver


@pytest.fixture(scope="function")
def mock_server(mocker):
    def get(url):
        r = mocker.MagicMock()
        if url.endswith("/nodes/"):
            r.json.return_value = {"code": 0, "result": ["IOMux-5a6ecbea"]}
        if "/pins/" in url:
            r.json.return_value = {"code": 0, "result": 1}
        return r

    def post(url, data=None):
        r = mocker.MagicMock()
        if "/pins/" in url:
            r.json.return_value = {"code": 0, "result": None}
        return r

    mock_get = mocker.patch("requests.get")
    mock_get.side_effect = get
    mock_post = mocker.patch("requests.post")
    mock_post.side_effect = post
    return (mock_get, mock_post)


@pytest.fixture(scope="function")
def lxa_pin(mock_server, target):
    r = LXAIOBusPIO(target, name=None, host="localhost:8080", node="IOMux-5a6ecbea", pin="OUT0")
    return r


@pytest.fixture(scope="function")
def lxa_driver(mock_server, target, lxa_pin):
    mock_get, mock_post = mock_server

    s = LXAIOBusPIODriver(target, name=None)
    target.activate(s)
    mock_get.assert_called_with("http://localhost:8080/nodes/")
    return s


def test_lxa_resource_instance(mock_server, lxa_pin):
    assert isinstance(lxa_pin, LXAIOBusPIO)


def test_lxa_driver_instance(mock_server, lxa_driver):
    assert isinstance(lxa_driver, LXAIOBusPIODriver)


def test_lxa_set(mock_server, lxa_driver):
    mock_get, mock_post = mock_server

    lxa_driver.set(True)
    mock_post.assert_called_once_with("http://localhost:8080/nodes/IOMux-5a6ecbea/pins/OUT0/", data={"value": "1"})


def test_lxa_unset(mock_server, lxa_driver):
    mock_get, mock_post = mock_server

    lxa_driver.set(False)
    mock_post.assert_called_once_with("http://localhost:8080/nodes/IOMux-5a6ecbea/pins/OUT0/", data={"value": "0"})


def test_lxa_get(mock_server, lxa_driver):
    mock_get, mock_post = mock_server
    val = lxa_driver.get()
    mock_get.assert_called_with("http://localhost:8080/nodes/IOMux-5a6ecbea/pins/OUT0/")
    assert val == True



#! /usr/bin/env python3
#
# Copyright 2021 Garmin Ltd. or its subsidiaries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import asyncio
from contextlib import contextmanager
from labgrid.remote.client import start_session
from labgrid.remote.generated import labgrid_coordinator_pb2
from labgrid.util.proxy import proxymanager
import os
import sys
import textwrap
import yaml


def main():
    @contextmanager
    def get_file(name, mode, standard):
        if name == "-":
            yield standard
        else:
            with open(name, mode) as f:
                yield f

    async def do_sync(session, args):
        with get_file(args.places, "r", sys.stdin) as f:
            config = yaml.safe_load(f)

        config.setdefault("places", {})

        changed = False
        seen_places = set()
        remove_places = set()
        for name, place in session.places.items():
            if name in config["places"]:
                seen_places.add(name)
            else:
                remove_places.add(name)

        for name in remove_places:
            print(f"Removing place {name}")
            if not args.dry_run:
                request = labgrid_coordinator_pb2.DeletePlaceRequest(name=name)
                await session.stub.DeletePlace(request)
                await session.sync_with_coordinator()

            changed = True

        for name in config["places"]:
            if not name in seen_places:
                print(f"Adding place {name}")
                if not args.dry_run:
                    request = labgrid_coordinator_pb2.AddPlaceRequest(name=name)
                    await session.stub.AddPlace(request)
                    await session.sync_with_coordinator()

                changed = True

        for name in config["places"]:
            matches = []
            for m in config["places"][name].get("matches", []):
                if isinstance(m, dict):
                    match = list(m.keys())[0]
                    matches.append((match, m[match]))
                else:
                    matches.append((m, None))

            seen_matches = set()
            remove_matches = set()
            place_tags = {}
            if name in seen_places:
                place = session.places[name]
                for m in [(repr(x), x.rename) for x in place.matches]:
                    if m in matches:
                        seen_matches.add(m)
                    else:
                        remove_matches.add(m)
                place_tags = place.tags

            for m in remove_matches:
                match, rename = m
                if rename:
                    print(f"Deleting named match '{match} -> {rename}' for place {name}")
                else:
                    print(f"Deleting match '{match}' for place {name}")
                if not args.dry_run:
                    request = labgrid_coordinator_pb2.DeletePlaceMatchRequest(placename=name, pattern=match)
                    await session.stub.DeletePlaceMatch(request)
                    await session.sync_with_coordinator()

                changed = True

            for m in matches:
                if not m in seen_matches:
                    match, rename = m
                    if rename:
                        print(f"Adding named match '{match} -> {rename}' for place {name}")
                    else:
                        print(f"Adding match '{match}' for place {name}")

                    if not args.dry_run:
                        request = labgrid_coordinator_pb2.AddPlaceMatchRequest(
                            placename=name, pattern=match, rename=rename
                        )
                        await session.stub.AddPlaceMatch(request)
                        await session.sync_with_coordinator()
                    changed = True

            tags = {str(k): str(v) for k, v in config["places"][name].get("tags", {}).items()}

            if place_tags != tags:
                print(
                    "Setting tags for place %s to %s"
                    % (
                        name,
                        ", ".join("%s=%s" % (key, value) for (key, value) in tags.items()),
                    )
                )

                # Set the empty string for tags that should be removed
                for k in place_tags:
                    if k not in tags:
                        tags[k] = ""

                if not args.dry_run:
                    request = labgrid_coordinator_pb2.SetPlaceTagsRequest(placename=name, tags=tags)
                    await session.stub.SetPlaceTags(request)
                    await session.sync_with_coordinator()

                changed = True

    async def do_dump(session, args):
        config = {"places": {}}
        for name, place in session.places.items():
            config["places"][name] = {
                "matches": [{repr(m): m.rename} if m.rename else repr(m) for m in place.matches],
                "tags": {k: v for k, v in place.tags.items()},
            }

        with get_file(args.dest, "w", sys.stdout) as f:
            yaml.dump(config, f)

    parser = argparse.ArgumentParser(
        description="Synchronize Labgrid places",
        epilog=textwrap.dedent(
            """\
            The YAML files describe what places should exist and what match
            strings and tags should be assigned to those places. The files are
            structured like:

              places: # A dictonary of places where each key is a place name
                my-place1: # Replace with your place
                  matches: # A list of match patterns. Replace with your match patterns
                    - "*/my-place1/*"
                    - "exporter/my-place1/resource": name # named matches supported
                  tags: # A dictionary of key/value tags. Replace with your tags
                    board: awesomesauce
                    bar: baz

            When syncing places, tags, and matches will be added or removed until the
            remote configuration matches the one in the YAML file
            """
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--coordinator",
        "-x",
        metavar="ADDRESS",
        default=os.environ.get("LG_COORDINATOR", "127.0.0.1:20408"),
        help="Coordinator address as HOST[:PORT] (default: %(default)s)",
    )
    parser.add_argument("--proxy", "-P", help="Proxy connections via given ssh host")

    subparsers = parser.add_subparsers()
    subparsers.required = True

    sync_parser = subparsers.add_parser("sync", help="Synchronize coordinator places with file")
    sync_parser.add_argument(
        "places",
        metavar="FILE",
        help="Places configuration YAML file. Use '-' for stdin",
    )
    sync_parser.add_argument(
        "--dry-run",
        "-n",
        action="store_true",
        help="Don't make any changes, only show what would be done",
    )
    sync_parser.set_defaults(func=do_sync)

    dump_parser = subparsers.add_parser(
        "dump",
        help="Dump existing places configuration to a YAML file. The dumped file is suitable for passing to `sync`",
    )
    dump_parser.add_argument(
        "dest",
        metavar="FILE",
        nargs="?",
        default="-",
        help="Destination file. Use '-' for stdout. Default is '%(default)s'",
    )
    dump_parser.set_defaults(func=do_dump)

    args = parser.parse_args()

    if args.proxy:
        proxymanager.force_proxy(args.proxy)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    session = start_session(
        args.coordinator,
        loop=loop,
    )

    try:
        return loop.run_until_complete(args.func(session, args))
    finally:
        loop.run_until_complete(session.stop())
        loop.run_until_complete(session.close())


if __name__ == "__main__":
    sys.exit(main())



import requests

PORT = 80


def power_set(host, port, index, value):
    index = int(index)
    assert 1 <= index <= 8

    if value:
        cgi = "ons.cgi"
    else:
        cgi = "offs.cgi"

    suffixstring = "0000000000000000"
    r = requests.get(
        f"http://{host}:{port}/{cgi}?led={1 << 8 - index:08b}{suffixstring}",
        auth=("snmp", "1234"),
    )
    r.raise_for_status()


def power_get(host, port, index):
    index = int(index)
    assert 1 <= index <= 8

    r = requests.get(
        f"http://{host}:{port}/status.xml",
        auth=("snmp", "1234"),
    )
    r.raise_for_status()

    state = r.text.split(",")[9 + index]
    return bool(int(state, 2))



import pytest

from labgrid.driver import ShellDriver, ExecutionError
from labgrid.exceptions import NoDriverFoundError

from ipaddress import IPv4Interface


class TestShellDriver:
    def test_instance(self, target, serial_driver):
        s = ShellDriver(target, "shell", "", "", "")
        assert isinstance(s, ShellDriver)

    def test_no_driver(self, target):
        with pytest.raises(NoDriverFoundError):
            ShellDriver(target, "shell", "", "", "")

    def test_run(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = ShellDriver(t, "shell", prompt="dummy", login_prompt="dummy", username="dummy")
        d.on_activate = mocker.MagicMock()
        d = t.get_driver("ShellDriver")
        d._run = mocker.MagicMock(return_value=(["success"], [], 0))
        res = d.run_check("test")
        assert res == ["success"]
        res = d.run("test")
        assert res == (["success"], [], 0)

    def test_run_error(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = ShellDriver(t, "shell", prompt="dummy", login_prompt="dummy", username="dummy")
        d.on_activate = mocker.MagicMock()
        d = t.get_driver("ShellDriver")
        d._run = mocker.MagicMock(return_value=(["error"], [], 1))
        with pytest.raises(ExecutionError):
            res = d.run_check("test")
        res = d.run("test")
        assert res == (["error"], [], 1)

    def test_run_with_timeout(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = ShellDriver(t, "shell", prompt="dummy", login_prompt="dummy", username="dummy")
        d.on_activate = mocker.MagicMock()
        d = t.get_driver("ShellDriver")
        d._run = mocker.MagicMock(return_value=(["success"], [], 0))
        res = d.run_check("test", timeout=30.0)
        assert res == ["success"]
        res = d.run("test")
        assert res == (["success"], [], 0)

    def test_default_interface_device_name(self, target_with_fakeconsole, mocker):
        fake_default_route_show = "default via 10.0.2.2 dev br-lan  src 10.0.2.15"

        t = target_with_fakeconsole
        d = ShellDriver(t, "shell", prompt="dummy", login_prompt="dummy", username="dummy")
        d.on_activate = mocker.MagicMock()
        d = t.get_driver("ShellDriver")
        d._run = mocker.MagicMock(return_value=([fake_default_route_show], [], 0))

        res = d.get_default_interface_device_name()
        assert res == "br-lan"

    def test_get_ip_addresses(self, target_with_fakeconsole, mocker):
        fake_ip_addr_show = r"""
18: br-lan.42    inet 192.168.42.1/24 brd 192.168.42.255 scope global br-lan.42\       valid_lft forever preferred_lft forever
18: br-lan.42    inet6 fe80::9683:c4ff:fea6:fb6b/64 scope link \       valid_lft forever preferred_lft forever
"""

        t = target_with_fakeconsole
        d = ShellDriver(t, "shell", prompt="dummy", login_prompt="dummy", username="dummy")
        d.on_activate = mocker.MagicMock()
        d = t.get_driver("ShellDriver")
        d._run = mocker.MagicMock(return_value=([fake_ip_addr_show], [], 0))

        res = d.get_ip_addresses("br-lan.42")
        assert res[0] == IPv4Interface("192.168.42.1/24")

    def test_get_ip_addresses_default(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = ShellDriver(t, "shell", prompt="dummy", login_prompt="dummy", username="dummy")
        d.on_activate = mocker.MagicMock()
        d = t.get_driver("ShellDriver")
        d._run = mocker.MagicMock()
        d._run.side_effect = [
            (["default via 192.168.42.255 dev br-lan.42  src 192.168.42.1"], [], 0),
            (
                [
                    r"""
18: br-lan.42    inet 192.168.42.1/24 brd 192.168.42.255 scope global br-lan.42\       valid_lft forever preferred_lft forever
18: br-lan.42    inet6 fe80::9683:c4ff:fea6:fb6b/64 scope link \       valid_lft forever preferred_lft forever
"""
                ],
                [],
                0,
            ),
        ]

        res = d.get_ip_addresses()
        assert res[0] == IPv4Interface("192.168.42.1/24")



import pytest


@pytest.fixture
def graph_strategy(target):
    from labgrid.strategy import GraphStrategy

    class TestStrategy(GraphStrategy):
        def state_Root(self):
            pass

        @GraphStrategy.depends("Root")
        def state_A1(self):
            pass

        @GraphStrategy.depends("Root")
        def state_A2(self):
            pass

        @GraphStrategy.depends("A1", "A2")
        def state_B(self):
            pass

        @GraphStrategy.depends("B")
        def state_C1(self):
            pass

        @GraphStrategy.depends("B")
        def state_C2(self):
            pass

        @GraphStrategy.depends("C1", "C2")
        def state_D(self):
            pass

    return TestStrategy(target, "strategy")


# api tests ###################################################################
@pytest.mark.dependency
def test_fixture(graph_strategy, target):
    # test root state
    assert graph_strategy.root_state == "Root"

    # test path
    assert graph_strategy.path == []

    # test state discovering
    state_names = sorted(list(graph_strategy.states.keys()))
    expected_state_names = ["A1", "A2", "B", "C1", "C2", "D", "Root"]

    assert state_names == expected_state_names


@pytest.mark.dependency(depends=["test_fixture"])
def test_strategy_without_states(target):
    from labgrid.strategy import GraphStrategy, InvalidGraphStrategyError

    class TestStrategy(GraphStrategy):
        pass

    with pytest.raises(InvalidGraphStrategyError) as e:
        TestStrategy(target, "strategy")

    assert e.value.msg == "GraphStrategies without states are invalid"


@pytest.mark.dependency(depends=["test_fixture"])
def test_strategy_without_root_state(target):
    from labgrid.strategy import GraphStrategy, InvalidGraphStrategyError

    class TestStrategy(GraphStrategy):
        @GraphStrategy.depends("B")
        def state_A(self):
            pass

        @GraphStrategy.depends("A")
        def state_B(self):
            pass

    with pytest.raises(InvalidGraphStrategyError) as e:
        TestStrategy(target, "strategy")

    assert e.value.msg == "GraphStrategies without root state are invalid"


@pytest.mark.dependency(depends=["test_fixture"])
def test_multiple_root_states(target):
    from labgrid.strategy import GraphStrategy, InvalidGraphStrategyError

    class TestStrategy(GraphStrategy):
        def state_Root(self):
            pass

        def state_Root2(self):
            pass

    with pytest.raises(InvalidGraphStrategyError) as e:
        TestStrategy(target, "strategy")

    assert e.value.msg.startswith("Only one root state supported.")


@pytest.mark.dependency(depends=["test_fixture"])
def test_unknown_dependencies(target):
    from labgrid.strategy import GraphStrategy, InvalidGraphStrategyError

    class TestStrategy(GraphStrategy):
        def state_Root(self):
            pass

        @GraphStrategy.depends("Root")
        def state_A(self):
            pass

        @GraphStrategy.depends("Root", "A", "C")
        def state_B(self):
            pass

    with pytest.raises(InvalidGraphStrategyError) as e:
        TestStrategy(target, "strategy")

    assert e.value.msg.startswith("B: State 'C' is unknown")


@pytest.mark.dependency(depends=["test_fixture"])
def test_strategy_with_uncallable_states(target):
    from labgrid.strategy import GraphStrategy, InvalidGraphStrategyError

    class TestStrategy(GraphStrategy):
        state_foo = 1

        def state_Root(self):
            pass

    with pytest.raises(InvalidGraphStrategyError) as e:
        TestStrategy(target, "strategy")

    assert e.value.msg.startswith("GraphStrategy state 'state_foo' is not callable")


@pytest.mark.dependency(
    name="api-works",
    depends=[
        "test_fixture",
        "test_strategy_without_states",
        "test_strategy_without_root_state",
        "test_unknown_dependencies",
        "test_strategy_with_uncallable_states",
    ],
)
def test_api_works():
    pass


# functional tests ############################################################
@pytest.mark.dependency(depends=["api-works"])
def test_graphviz_graph(graph_strategy):
    pytest.importorskip("graphviz")
    graph_strategy.graph
    graph_strategy.graph  # trigger the caching branch in the graph code

    graph_strategy.path = ["Root", "A1", "B"]  # fake a post transition graph
    graph_strategy.graph


@pytest.mark.dependency(depends=["api-works"])
def test_transition(graph_strategy):
    assert graph_strategy.transition("B") == ["Root", "A1", "B"]
    assert graph_strategy.path == ["Root", "A1", "B"]

    assert graph_strategy.transition("B") == []
    assert graph_strategy.path == ["Root", "A1", "B"]


@pytest.mark.dependency(depends=["api-works", "test_transition"])
def test_transition_to_unknown_state(graph_strategy):
    from labgrid.strategy import GraphStrategyRuntimeError

    with pytest.raises(GraphStrategyRuntimeError) as e:
        graph_strategy.transition("G")

    assert e.value.msg.startswith("Unknown state 'G'.")


@pytest.mark.dependency(depends=["api-works", "test_transition"])
def test_interleaved_transitions(target):
    from labgrid.strategy import GraphStrategy, GraphStrategyRuntimeError

    class TestStrategy(GraphStrategy):
        def state_Root(self):
            pass

        @GraphStrategy.depends("Root")
        def state_A(self):
            self.transition("B")

        @GraphStrategy.depends("Root")
        def state_B(self):
            self.transition("A")

    strategy = TestStrategy(target, "strategy")

    with pytest.raises(GraphStrategyRuntimeError) as e:
        strategy.transition("A")

    assert e.value.msg == "Another transition is already running"


@pytest.mark.dependency(depends=["api-works", "test_transition"])
def test_transition_via(graph_strategy):
    assert graph_strategy.transition("D", via=["A2"]) == ["Root", "A2", "B", "C1", "D"]
    assert graph_strategy.path == ["Root", "A2", "B", "C1", "D"]

    graph_strategy.invalidate()

    assert graph_strategy.transition("D:A2") == ["Root", "A2", "B", "C1", "D"]
    assert graph_strategy.path == ["Root", "A2", "B", "C1", "D"]

    assert graph_strategy.transition("D:C2,A2") == ["Root", "A2", "B", "C2", "D"]
    assert graph_strategy.path == ["Root", "A2", "B", "C2", "D"]


@pytest.mark.dependency(depends=["api-works", "test_transition", "test_transition_via"])
def test_incremental_transition(graph_strategy):
    assert graph_strategy.transition("B") == ["Root", "A1", "B"]
    assert graph_strategy.transition("D") == ["C1", "D"]

    graph_strategy.invalidate()

    assert graph_strategy.transition("B", via=["A2"]) == ["Root", "A2", "B"]
    assert graph_strategy.transition("D") == ["Root", "A1", "B", "C1", "D"]


@pytest.mark.dependency(depends=["api-works", "test_transition"])
def test_transition_error(target):
    from labgrid.strategy import GraphStrategy

    class TestStrategy(GraphStrategy):
        def state_Root(self):
            pass

        @GraphStrategy.depends("Root")
        def state_A(self):
            pass

        @GraphStrategy.depends("A")
        def state_B(self):
            raise Exception

    strategy = TestStrategy(target, "strategy")

    strategy.transition("A")
    assert strategy.path == ["Root", "A"]

    with pytest.raises(Exception):
        strategy.transition("B")

    with pytest.raises(Exception):
        strategy.transition("B", via="B")

    assert strategy.path == []



from .step import steps


class StepReporter:
    _started = False

    def __init__(self):
        from warnings import warn

        warn(
            "StepReporter should not be instantiated, use StepReporter.start()/.stop() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    @classmethod
    def start(cls):
        """starts the StepReporter"""
        from warnings import warn

        warn(
            """
            StepReporter is deprecated, use the StepLogger and basicConfig from labgrid.logging
            instead which integrates with the python logging infrastructure.
            """,
            DeprecationWarning,
            stacklevel=2,
        )
        assert not cls._started
        steps.subscribe(cls.notify)
        cls._started = True

    @classmethod
    def stop(cls):
        """stops the StepReporter"""
        assert cls._started
        steps.unsubscribe(cls.notify)
        cls._started = False

    @staticmethod
    def notify(event):
        # ignore tagged events
        if event.step.tag:
            return

        step = event.step
        indent = "  " * step.level
        print(f"{indent}{event}")



import logging
import subprocess
import sys
from time import time
import attr

from ..factory import target_factory
from .common import ManagedResource, ResourceManager

@attr.s
class SNMPSwitch:
    """SNMPSwitch describes a switch accessible over SNMP. This class
    implements functions to query ports and the forwarding database."""
    hostname = attr.ib(validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        self.logger = logging.getLogger(f"{self}")
        self.ports = {}
        self.fdb = {}
        self.macs_by_port = {}
        self._autodetect()

    def _autodetect(self):
        from pysnmp import hlapi

        for (errorIndication, errorStatus, _, varBindTable) in hlapi.getCmd(
                hlapi.SnmpEngine(),
                hlapi.CommunityData('public'),
                hlapi.UdpTransportTarget((self.hostname, 161)),
                hlapi.ContextData(),
                hlapi.ObjectType(hlapi.ObjectIdentity('SNMPv2-MIB', 'sysDescr', 0))):
            if errorIndication:
                raise Exception(f"snmp error {errorIndication}")
            elif errorStatus:
                raise Exception(f"snmp error {errorStatus}")
            else:
                sysDescr = str(varBindTable[0][1])

        if sysDescr.startswith("HPE OfficeConnect Switch 1820 24G J9980A,"):
            self._get_fdb = self._get_fdb_dot1q
        elif sysDescr.startswith("HP 1810-24G,"):
            self._get_fdb = self._get_fdb_dot1d
        else:
            raise Exception(f"unsupported switch {sysDescr}")

        self.logger.debug("autodetected switch%s: %s %s", sysDescr, self._get_ports, self._get_fdb)

    def _get_ports(self):
        """Fetch ports and their values via SNMP

        Returns:
            Dict[Dict[]]: ports and their values
        """
        from pysnmp import hlapi

        variables = [
            (hlapi.ObjectType(hlapi.ObjectIdentity('IF-MIB', 'ifIndex')), 'index'),
            (hlapi.ObjectType(hlapi.ObjectIdentity('IF-MIB', 'ifDescr')), 'descr'),
            (hlapi.ObjectType(hlapi.ObjectIdentity('IF-MIB', 'ifSpeed')), 'speed'),
            (hlapi.ObjectType(hlapi.ObjectIdentity('IF-MIB', 'ifOperStatus')), 'status'),
            (hlapi.ObjectType(hlapi.ObjectIdentity('IF-MIB', 'ifInErrors')), 'inErrors'),
            (hlapi.ObjectType(hlapi.ObjectIdentity('IF-MIB', 'ifHCInOctets')), 'inOctets'),
            (hlapi.ObjectType(hlapi.ObjectIdentity('IF-MIB', 'ifHCOutOctets')), 'outOctets'),
        ]
        ports = {}

        for (errorIndication, errorStatus, _, varBindTable) in hlapi.bulkCmd(
                hlapi.SnmpEngine(),
                hlapi.CommunityData('public'),
                hlapi.UdpTransportTarget((self.hostname, 161)),
                hlapi.ContextData(),
                0, 20,
                *[x[0] for x in variables],
                lexicographicMode=False):
            if errorIndication:
                raise Exception(f"snmp error {errorIndication}")
            elif errorStatus:
                raise Exception(f"snmp error {errorStatus}")
            else:
                port = {}
                for (_, val), (_, label) in zip(varBindTable, variables):
                    val = val.prettyPrint()
                    if label == 'status':
                        val = val.strip("'")
                    port[label] = val
                ports[port.pop('index')] = port

        return ports

    def _get_fdb_dot1d(self):
        """Fetch the forwarding database via SNMP using the BRIDGE-MIB

        Returns:
            Dict[List[str]]: ports and their values
        """
        from pysnmp import hlapi

        ports = {}

        for (errorIndication, errorStatus, _, varBindTable) in hlapi.bulkCmd(
                hlapi.SnmpEngine(),
                hlapi.CommunityData('public'),
                hlapi.UdpTransportTarget((self.hostname, 161)),
                hlapi.ContextData(),
                0, 50,
                hlapi.ObjectType(hlapi.ObjectIdentity('BRIDGE-MIB', 'dot1dTpFdbPort')),
                lexicographicMode=False):
            if errorIndication:
                raise Exception(f"snmp error {errorIndication}")
            elif errorStatus:
                raise Exception(f"snmp error {errorStatus}")
            else:
                for varBinds in varBindTable:
                    key, val = varBinds
                    if not val:
                        continue
                    mac = key.getMibSymbol()[-1][0].prettyPrint()
                    interface = str(int(val))
                    ports.setdefault(interface, []).append(mac)

        return ports

    def _get_fdb_dot1q(self):
        """Fetch the forwarding database via SNMP using the Q-BRIDGE-MIB

        Returns:
            Dict[List[str]]: ports and their values
        """
        from pysnmp import hlapi

        ports = {}

        for (errorIndication, errorStatus, _, varBindTable) in hlapi.bulkCmd(
                hlapi.SnmpEngine(),
                hlapi.CommunityData('public'),
                hlapi.UdpTransportTarget((self.hostname, 161)),
                hlapi.ContextData(),
                0, 50,
                hlapi.ObjectType(hlapi.ObjectIdentity('Q-BRIDGE-MIB', 'dot1qTpFdbPort')),
                lexicographicMode=False):
            if errorIndication:
                raise Exception(f"snmp error {errorIndication}")
            elif errorStatus:
                raise Exception(f"snmp error {errorStatus}")
            else:
                for varBinds in varBindTable:
                    key, val = varBinds
                    if not val:
                        continue
                    mac = key.getMibSymbol()[-1][1].prettyPrint()
                    interface = str(int(val))
                    ports.setdefault(interface, []).append(mac)

        return ports

    def _update_macs(self):
        """remember the first time we've seen a MAC on a port"""
        for removed in self.macs_by_port.keys() - self.ports.keys():
            del self.macs_by_port[removed]
        for interface, macs in self.fdb.items():
            seen = self.macs_by_port.setdefault(interface, {})
            for removed in seen.keys() - set(macs):
                del seen[removed]
            for mac in macs:
                seen.setdefault(mac, int(time()))

    def update(self):
        """Update port status and forwarding database status

        Returns:
            None
        """
        self.logger.debug("polling switch FDB")
        self.fdb = self._get_fdb()
        self.logger.debug("polling switch ports")
        self.ports = self._get_ports()
        self.logger.debug("updating macs by port")
        self._update_macs()


@attr.s
class EthernetPortManager(ResourceManager):
    """The EthernetPortManager periodically polls the switch for new updates."""
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.loop = None
        self.poll_tasks = []
        self.switches = {}
        self.neighbors = {}

    def on_resource_added(self, resource):
        """Handler to execute when the resource is added

        Checks whether the resource can be managed by this Manager and starts
        the event loop.

        Args:
            resource(Resource): resource to check against

        Returns:
            None
        """
        if not isinstance(resource, SNMPEthernetPort):
            return
        self._start()
        resource.avail = True

    def _start(self):
        """Internal function to register as task and attach/start the event
        loop

        Returns:
            None
        """
        import asyncio

        if self.poll_tasks:
            return

        async def poll_neighbour(self):
            self.logger.debug("polling neighbor table")
            self.neighbors = EthernetPortManager._get_neigh()

            await asyncio.sleep(1.0)

        async def poll_switches(self):
            current = set(resource.switch for resource in self.resources)
            removed = set(self.switches) - current
            new = current - set(self.switches)
            for switch in removed:
                del self.switches[switch]
            for switch in new:
                self.switches[switch] = SNMPSwitch(switch)
            for switch in current:
                self.switches[switch].update()
                await asyncio.sleep(1.0)

            await asyncio.sleep(2.0)

        async def poll(self, handler):
            while True:
                try:
                    await asyncio.sleep(1.0)
                    await handler(self)
                except asyncio.CancelledError:
                    break
                except Exception:  # pylint: disable=broad-except
                    import traceback
                    traceback.print_exc(file=sys.stderr)

        self.loop = asyncio.get_event_loop()
        self.poll_tasks.append(self.loop.create_task(poll(self, poll_neighbour)))
        self.poll_tasks.append(self.loop.create_task(poll(self, poll_switches)))

    @staticmethod
    def _get_neigh():
        """Internal function to retrieve the neighbors on the test machine

        Returns:
            Dict[Tuple[(str,str,str)]]: dictionary with mac addresses as keys
            and a Tuple of address, device and state as values
        """
        neighbors = {}

        for line in subprocess.check_output(['ip', 'neigh', 'show']).splitlines():
            line = line.decode('ascii').strip().split()
            addr = line.pop(0)
            if line[0] == 'dev':
                line.pop(0) # "dev"
                line.pop(0) # actual dev
            if line[0] == 'lladdr':
                line.pop(0)
                lladdr = line.pop(0)
            else:
                lladdr = None
            if line[0] == 'router':
                line.pop()
            line.pop(0) # state
            assert not line
            # TODO: check if we could use the device and state information
            neighbors.setdefault(lladdr, []).append(addr)
        for value in neighbors.values():
            value.sort()

        return neighbors

    def poll(self):
        """Updates the state with new information from the event loop

        Returns:
            None
        """
        import asyncio
        if not self.loop.is_running():
            self.loop.run_until_complete(asyncio.sleep(0.0))
        for resource in self.resources:
            switch = self.switches.get(resource.switch)
            if not switch:
                resource.extra = None
                continue
            extra = {}
            for mac, timestamp in switch.macs_by_port.get(resource.interface, {}).items():
                extra.setdefault('macs', {})[mac] = {
                    'timestamp': timestamp,
                    'ips': self.neighbors.get(mac, []),
                }
            extra.update(switch.ports.get(resource.interface, {}))
            if resource.extra != extra:
                resource.extra = extra
                self.logger.debug("new information for %s: %s", resource, extra)


@target_factory.reg_resource
@attr.s
class SNMPEthernetPort(ManagedResource):
    """SNMPEthernetPort describes an ethernet port which can be queried over
    SNMP.

    Args:
        switch (str): hostname of the switch to query
        interface (str): name of the interface to query
    """
    manager_cls = EthernetPortManager

    switch = attr.ib(validator=attr.validators.instance_of(str))
    interface = attr.ib(validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.extra = {}



import re
import requests

from ..exception import ExecutionError

# This driver implementes a power port for the EG_PMS2_LAN & EG_PMS2_WLAN
# devices, it works through a simple HTTP interface, that requires a login.
# Driver has been tested with:
# * EG_PMS2_LAN

# The default HTTP port for usage via the SSH proxy.
PORT = 80

# The login was successful when we can locate a Log out button
LOGIN_SUCCESS_REGEX = r'<div class="boxmenuitem"><a href="login\.html">Log Out</a></div>'
# Search for a string similar to: "var sockstates = [1,1,1,1];" and create
# a match group out of the numbers within the square brackets.
SOCKSTATES_REGEX = r"var\s+sockstates\s+=\s+\[([01]),([01]),([01]),([01])\];"


def login(base_url: str) -> None:
    """
    Use the default password 1, because labgrid doesn't support password
    encryption, modifying the password doesn't secure the device as the
    password would be stored as plain-text.
    """
    login_url = f"{base_url}/login.html"
    try:
        response = requests.post(login_url, data={"pw": 1})
    except requests.exceptions.ConnectionError as err:
        raise ExecutionError(
            f"Device not found at {base_url} or the network interface of the "
            "device is not active (press the reset button on the device)"
        ) from err
    if response.status_code != 200 or not re.search(LOGIN_SUCCESS_REGEX, response.text):
        raise ExecutionError("Login to Energenie web interface failed")


def logout(base_url: str) -> None:
    """
    After a successful login, the session is reserved for the IP address.
    Logout explicitly to allow accessing the device from different hosts.
    """
    response = requests.get(f"{base_url}/login.html")
    if response.status_code != 200:
        raise ExecutionError("Logout from Energenie web interface failed")


def power_set(host, port, index, value):
    base_url = f"http://{host}:{port}"
    index = int(index)
    assert 1 <= index <= 4

    value = 1 if value else 0
    login(base_url=base_url)
    response = requests.post(base_url, data={f"cte{index}": value})
    response.raise_for_status()
    logout(base_url=base_url)
    if not response.status_code == 200:
        raise ExecutionError(f"Switching socket at index {index} {'on' if value else 'off'} failed")


def power_get(host, port, index):
    base_url = f"http://{host}:{port}"
    index = int(index)
    assert 1 <= index <= 4

    # Fetch status
    login(base_url=base_url)
    response = requests.get(f"{base_url}/energenie.html")
    response.raise_for_status()
    logout(base_url=base_url)
    match_group = re.search(SOCKSTATES_REGEX, response.text)
    sockstates = [int(match_group.group(x)) for x in range(1, 5)]
    return sockstates[index - 1] == 1



def test_usb(target):
    r = target.get_resource("MXSUSBLoader")
    assert r.devnum is not None
    assert r.busnum is not None


def test_mxs_load(target):
    bp = target.get_driver("BootstrapProtocol")
    bp.load()



import inspect
import os
import warnings
from functools import wraps
from time import monotonic


# TODO: collect events from all Steps and combine when possible, only flush
# after some time
class Steps:
    def __init__(self):
        self._stack = []
        self._subscribers = []

    def get_current(self):
        return self._stack[-1] if self._stack else None

    def get_new(self, title, tag, source, sourceinfo):
        step = Step(title, level=len(self._stack) + 1, tag=tag, source=source, sourceinfo=sourceinfo)  # pylint: disable=redefined-outer-name
        return step

    def push(self, step):  # pylint: disable=redefined-outer-name
        assert step not in self._stack
        self._stack.append(step)
        step.parent = self.get_current()
        step.level = len(self._stack)

    def pop(self, step):  # pylint: disable=redefined-outer-name
        assert self._stack[-1] is step
        self._stack.pop()

    def subscribe(self, callback):
        self._subscribers.append(callback)

    def unsubscribe(self, callback):
        assert callback in self._subscribers
        self._subscribers.remove(callback)

    def notify(self, event):
        # TODO: buffer and try to merge consecutive events
        for subscriber in self._subscribers:
            try:
                subscriber(event)
            except Exception as e:  # pylint: disable=broad-except
                warnings.warn(f"unhandled exception during event notification: {e}")


steps = Steps()


class StepEvent:
    def __init__(self, step, data, *, resource=None, stream=False):  # pylint: disable=redefined-outer-name
        self.ts = monotonic()  # used to keep track of the events age
        self.step = step
        self.data = data
        self.resource = resource
        self.stream = stream

    def __str__(self):
        result = [self.step.title]
        if self.resource:
            result.append(self.resource.__class__.__name__)
        data = self.data.copy()
        duration = data.pop("duration", 0.0)
        pairs = [f"{k}={repr(v)}" for k, v in data.items() if v is not None]
        if duration >= 0.001:
            pairs.append(f"duration={duration:.3f}")
        result.append(", ".join(pairs))
        return " ".join(result)

    def __setitem__(self, k, v):
        self.data[k] = v

    def _invalidate(self):
        self.ts = None
        self.step = None
        self.data = None
        self.resource = None
        self.stream = None

    def merge(self, other):
        if not self.stream and not other.stream:
            return False
        if self.ts > other.ts:
            return False
        if self.resource is not other.resource:
            return False
        if self.data.keys() != other.data.keys():
            return False
        for k, v in other.data:
            self.data[k] += v
        other._invalidate()
        return True

    @property
    def age(self):
        return monotonic() - self.ts


# TODO: allow attaching log information, using a Resource as meta-data
class Step:
    def __init__(self, title, level, tag, source, sourceinfo):
        self.title = title
        self.level = level
        self.tag = tag
        self.source = source
        self.sourceinfo = sourceinfo
        self.args = None
        self.result = None
        self.exception = None
        self._start_ts = None
        self._stop_ts = None
        self._skipped = False

    def __repr__(self):
        result = [f"Step(title={self.title!r}, level={self.level}, status={self.status}"]
        if self.args is not None:
            result.append(f", args={self.args}")
        if self.exception is not None:
            result.append(f", exception={self.exception}")
        if self.result is not None:
            result.append(f", result={self.result}")
        duration = self.duration
        if duration >= 0.001:
            result.append(f", duration={duration:.3f}")
        result.append(")")
        return "".join(result)

    @property
    def duration(self):
        if self._start_ts is None:
            return 0.0
        if self._stop_ts is None:
            return monotonic() - self._start_ts

        return self._stop_ts - self._start_ts

    @property
    def status(self):
        if self._start_ts is None:
            return "new"
        if self._stop_ts is None:
            return "active"

        return "done"

    @property
    def is_active(self):
        return self.status == "active"

    @property
    def is_done(self):
        return self.status == "done"

    def _notify(self, event: StepEvent):
        assert event.step is self
        steps.notify(event)

    def start(self):
        assert self._start_ts is None
        self._start_ts = monotonic()
        steps.push(self)
        self._notify(
            StepEvent(
                self,
                {
                    "state": "start",
                    "args": self.args,
                },
            )
        )

    def skip(self, reason):
        assert self._start_ts is not None
        self._notify(StepEvent(self, {"skip": reason}))

    def stop(self):
        assert self._start_ts is not None
        assert self._stop_ts is None
        self._stop_ts = monotonic()
        event = StepEvent(self, {"state": "stop"})
        if self.exception:
            event["exception"] = self.exception
        else:
            event["result"] = self.result
        duration = self.duration
        if duration:
            event["duration"] = duration
        self._notify(event)
        steps.pop(self)

    def __del__(self):
        if not self.is_done:
            warnings.warn(f"__del__ called before {step} was done")


def step(*, title=None, args=[], result=False, tag=None):
    def decorator(func):
        # resolve default title
        nonlocal title
        title = title or func.__name__

        signature = inspect.signature(func)

        @wraps(func)
        def wrapper(*_args, **_kwargs):
            bound = signature.bind_partial(*_args, **_kwargs)
            bound.apply_defaults()
            source = func.__self__ if inspect.ismethod(func) else bound.arguments.get("self")
            pathname = func.__code__.co_filename
            sourceinfo = (pathname, os.path.basename(pathname), func.__code__.co_firstlineno)
            step = steps.get_new(title, tag, source, sourceinfo)  # pylint: disable=redefined-outer-name
            # optionally pass the step object
            if "step" in signature.parameters:
                _kwargs["step"] = step
            if args:
                step.args = {k: bound.arguments[k] for k in args}
            step.start()
            try:
                _result = func(*_args, **_kwargs)
                if result:
                    step.result = _result
            except Exception as e:
                step.exception = e
                raise
            finally:
                step.stop()
            return _result

        wrapper.__signature__ = signature.replace(
            parameters=[p for p in signature.parameters.values() if p.name != "step"]
        )
        return wrapper

    return decorator



import abc


class CommandProtocol(abc.ABC):
    """Abstract class for the CommandProtocol"""

    @abc.abstractmethod
    def run(self, command: str):
        """
        Run a command
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_check(self, command: str):
        """
        Run a command, return str if successful, ExecutionError otherwise
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_status(self):
        """
        Get status of the Driver
        """
        raise NotImplementedError

    @abc.abstractmethod
    def wait_for(self):
        """
        Wait for a shell command to return with the specified output
        """
        raise NotImplementedError

    @abc.abstractmethod
    def poll_until_success(self):
        """
        Repeatedly call a shell command until it succeeds
        """
        raise NotImplementedError



import os.path
import attr

from ..resource import NetworkFlashrom
from ..factory import target_factory
from ..step import step
from ..protocol import BootstrapProtocol
from .common import Driver, check_file
from ..util.managedfile import ManagedFile
from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class FlashromDriver(Driver, BootstrapProtocol):
    """The Flashrom driver used the flashrom utility to write an image to a raw rom.
    The driver is a pure wrapper of the flashrom utility"""

    bindings = {
        "flashrom_resource": {"Flashrom", NetworkFlashrom},
    }

    image = attr.ib(default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool("flashrom")
        else:
            self.tool = "flashrom"
        self.logger.debug("Tool %s", self.tool)

    def _get_flashrom_prefix(self):
        return self.flashrom_resource.command_prefix + [self.tool]

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(title="call", args=["args"])
    def __call__(self, *args):
        arg_list = list(args)
        arg_list.append("-p")
        arg_list.append(f"{self.flashrom_resource.programmer}")
        processwrapper.check_output(self._get_flashrom_prefix() + arg_list)

    @Driver.check_active
    @step(args=["filename"])
    def load(self, filename=None):
        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        filename = os.path.abspath(filename)
        check_file(filename)
        mf = ManagedFile(filename, self.flashrom_resource)
        mf.sync_to_resource()
        self.logger.debug("Local File %s synced to remote path %s", filename, mf.get_remote_path())
        self("-w", mf.get_remote_path())



import pexpect
import pytest
import time
import re


@pytest.fixture
def power_env(tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """
targets:
  main:
    drivers:
      ManualPowerDriver: {}
"""
    )
    return p


@pytest.fixture
def pw_cycle_test(tmpdir):
    t = tmpdir.join("pw_cycle_test.py")
    t.write(
        """
def test(target):
    pw = target.get_driver("ManualPowerDriver")
    pw.cycle()
"""
    )
    return t


def test_step_logger(power_env, pw_cycle_test):
    # step reporter is called when -vv is given
    # -s is necessary for manual power driver confirmation
    with pexpect.spawn("pytest -vvs --lg-env {env} {test}".format(env=power_env, test=pw_cycle_test)) as spawn:
        # rough match
        spawn.expect("→.*?ManualPowerDriver.*?cycle.*?\r\n".encode("utf-8"))
        step_line = spawn.after.decode("utf-8").rstrip()
        # exact match
        assert step_line.endswith("→ ManualPowerDriver.cycle()"), f"'{step_line}' did not match"

        spawn.expect("main: CYCLE the target main and press enter")
        time.sleep(0.02)  # ensure that the step measures a duration
        spawn.sendline()

        # rough match
        spawn.expect("←.*?ManualPowerDriver.*?cycle.*?\r\n".encode("utf-8"))
        step_line = spawn.after.decode("utf-8").rstrip()
        # exact match

        assert re.match(r"← ManualPowerDriver.cycle\(\) \[[\d.]+s\]$", step_line), f"'{step_line}' did not match"

        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0



import pytest


@pytest.fixture()
def pdu(target):
    return target.get_driver("NetworkPowerDriver")


def test_something(pdu):
    pdu.on()
    # Note that the pdu might not have switched on when requesting it's
    # status immediately
    assert pdu.get() is True

    # ... do something

    pdu.off()
    assert pdu.get() is False



#! /usr/bin/env python3
#
# Copyright 2022 Garmin Ltd. or its subsidiaries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# All metrics are optionally prefixed by a set of tag values which can be
# specified with the --tag command line option.
#
# This program will connect to a labgrid coordinator and report the following
# stats about places to a statsd server on a periodic basis:
#
#   labgrid.places.[TAG[.TAG]].acquired (gauge): The number of places that are
#                                                currently acquired with this
#                                                set of tags
#   labgrid.places.[TAG[.TAG]].total (gauge): The number of places attached to
#                                             the coordinator with this set of
#                                             tags
#
# It also can report metrics about reservations with the following format:
#
#   labgrid.reservations.GROUP.[TAG[.TAG]].STATE (gauge): The number of
#                                                         reservations in the
#                                                         given STATE
#
# where:
#   GROUP is the name of the filter group used for the reservation. Currently,
#   just "main"
#
#   STATE is the state of the reservation, e.g. "acquired", "waiting", etc.


import sys
import argparse
import os
import time
import asyncio

from labgrid.remote.client import start_session, Error
from labgrid.remote.generated import labgrid_coordinator_pb2
from labgrid.remote.common import Reservation
import statsd


def inc_gauge(gauges, key):
    gauges.setdefault(key, 0)
    gauges[key] += 1


async def report_reservations(session, tags, gauges):
    request = labgrid_coordinator_pb2.GetReservationsRequest()

    response = await session.stub.GetReservations(request)
    reservations = [Reservation.from_pb2(x) for x in response.reservations]

    for reservation in reservations:
        groups = reservation.filters

        if not groups:
            groups = {"": {}}

        for group_name, group in groups.items():
            inc_gauge(
                gauges,
                ".".join(["reservations", group_name] + [group.get(t, "") for t in tags] + [reservation.state.name]),
            )


async def report_places(session, tags, gauges):
    acquired_count = 0
    total_count = 0

    for name, place in session.places.items():
        prefix = ".".join(["places"] + [place.tags.get(t, "") for t in tags])
        inc_gauge(gauges, f"{prefix}.total")
        if place.acquired:
            inc_gauge(gauges, f"{prefix}.acquired")


def main():
    parser = argparse.ArgumentParser(description="Report Labgrid usage metrics to statsd")
    parser.add_argument(
        "-x",
        "--coordinator",
        metavar="ADDRESS",
        help="Coordinator address as HOST[:PORT]. Default is %(default)s",
        default=os.environ.get("LG_COORDINATOR", "127.0.0.1:20408"),
    )
    parser.add_argument(
        "--period",
        help="How often to report reservations stats to statsd. Default is %(default)s",
        type=int,
        default=60,
    )
    parser.add_argument(
        "--statsd-protocol",
        metavar="PROTOCOL",
        choices=("udp", "tcp"),
        default="udp",
        help="Use specified protocol for statsd",
    )
    parser.add_argument(
        "--statsd-prefix",
        metavar="PREFIX",
        help="Prefix for statsd stats",
        default="labgrid",
    )
    parser.add_argument(
        "--statsd-server",
        metavar="HOST",
        help="Statsd server",
        default="localhost",
    )
    parser.add_argument(
        "--statsd-port",
        metavar="PORT",
        help="Statsd server port",
        type=int,
        default=8125,
    )
    parser.add_argument(
        "--tag",
        metavar="TAG",
        dest="tags",
        help="Tag values to include in statsd path for reservations. Repeat to include multiple tags",
        action="append",
        default=[],
    )

    args = parser.parse_args()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    statsd_client = None
    gauges = {}

    next_time = time.monotonic()

    while True:
        if statsd_client is None:
            if args.statsd_protocol == "udp":
                statsd_client = statsd.StatsClient(
                    host=args.statsd_server,
                    port=args.statsd_port,
                    prefix=args.statsd_prefix,
                )
            elif args.statsd_protocol == "tcp":
                statsd_client = statsd.TCPStatsClient(
                    host=args.statsd_server,
                    port=args.statsd_port,
                    prefix=args.statsd_prefix,
                )

        # Reset all known gauges to 0
        for key in gauges:
            gauges[key] = 0

        sleep_time = next_time - time.monotonic()
        if sleep_time > 0:
            time.sleep(sleep_time)

        next_time = time.monotonic() + args.period
        try:
            session = start_session(args.coordinator, loop=loop)
            try:
                loop.run_until_complete(
                    asyncio.gather(
                        report_places(session, args.tags, gauges),
                        report_reservations(session, args.tags, gauges),
                    )
                )
            finally:
                loop.run_until_complete(session.stop())
                loop.run_until_complete(session.close())
        except Error as e:
            print(f"Error communicating with labgrid: {e}")
            continue

        try:
            with statsd_client.pipeline() as pipe:
                for k, v in gauges.items():
                    pipe.gauge(k, v)
        except OSError as e:
            print(f"Error communication with statsd server: {e}")
            statsd_client = None
            continue


if __name__ == "__main__":
    sys.exit(main())



""" Tested with TP Link KP303, and should be compatible with any strip supported by kasa """

import asyncio
from kasa.iot import IotStrip


async def _power_set(host, port, index, value):
    """We embed the coroutines in an `async` function to minimise calls to `asyncio.run`"""
    assert port is None
    index = int(index)
    strip = IotStrip(host)
    await strip.update()
    assert (
        len(strip.children) > index
    ), "Trying to access non-existant plug socket on strip"
    if value is True:
        await strip.children[index].turn_on()
    elif value is False:
        await strip.children[index].turn_off()


def power_set(host, port, index, value):
    asyncio.run(_power_set(host, port, index, value))


def power_get(host, port, index):
    assert port is None
    index = int(index)
    strip = IotStrip(host)
    asyncio.run(strip.update())
    assert (
        len(strip.children) > index
    ), "Trying to access non-existant plug socket on strip"
    return strip.children[index].is_on



import attr


@attr.s(eq=False)
class ExecutionError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))
    stdout = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(list)))
    stderr = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(list)))


@attr.s(eq=False)
class CleanUpError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))



from collections import OrderedDict
from copy import deepcopy

import attr
import pytest

from labgrid import Target, target_factory
from labgrid.driver import Driver
from labgrid.resource import Resource
from labgrid.exceptions import InvalidConfigError, RegistrationError
from labgrid.resource import SerialPort
from labgrid.util.yaml import load

def test_empty():
    t = target_factory.make_target('dummy', {})
    assert isinstance(t, Target)


def test_resources():
    original_config = {
        'resources': OrderedDict([
            ('RawSerialPort', {
                'port': 'foo',
                'speed': 115200,
                'name': 'console',
            }),
        ]),
    }
    config = deepcopy(original_config)
    t = target_factory.make_target('dummy', config)
    assert isinstance(t, Target)
    assert t.get_resource(SerialPort) is not None
    assert config == original_config


def test_drivers():
    original_config = {
        'resources': OrderedDict([
            ('RawSerialPort', {
                'port': 'foo',
                'speed': 115200
            }),
        ]),
        'drivers': OrderedDict([
            ('FakeConsoleDriver', {
                'name': 'console',
            }),
            ('ShellDriver', {
                'name': 'shell',
                'prompt': '',
                'login_prompt': '',
                'username': ''
            }),
        ]),
    }
    config = deepcopy(original_config)
    t = target_factory.make_target('dummy', config)
    assert isinstance(t, Target)
    assert t.get_resource(SerialPort) is not None
    assert config == original_config


def test_convert_dict():
    original_data = load("""
    FooPort: {}
    BarPort:
      name: bar
    """)
    data = deepcopy(original_data)
    l = target_factory._convert_to_named_list(data)
    assert l == [
        {
            'cls': 'FooPort',
            'name': None,
        },
        {
            'cls': 'BarPort',
            'name': 'bar'
        },
    ]
    assert data == original_data


def test_convert_simple_list():
    original_data = load("""
    - FooPort: {}
    - BarPort:
        name: bar
    """)
    data = deepcopy(original_data)
    l = target_factory._convert_to_named_list(data)
    assert l == [
        {
            'cls': 'FooPort',
            'name': None,
        },
        {
            'cls': 'BarPort',
            'name': 'bar'
        },
    ]
    assert data == original_data


def test_convert_explicit_list():
    original_data = load("""
    - cls: FooPort
    - cls: BarPort
      name: bar
    """)
    data = deepcopy(original_data)
    l = target_factory._convert_to_named_list(data)
    assert l == [
        {
            'cls': 'FooPort',
            'name': None,
        },
        {
            'cls': 'BarPort',
            'name': 'bar'
        },
    ]
    assert data == original_data


def test_normalize_config():
    original_config = {
        'resources': OrderedDict([
            ('RawSerialPort', {
                'port': 'foo',
                'speed': 115200
            }),
        ]),
        'drivers': OrderedDict([
            ('FakeConsoleDriver', {
                'name': 'console',
            }),
        ]),
    }
    config = deepcopy(original_config)
    resources, drivers = target_factory.normalize_config(config)

    assert 'RawSerialPort' in resources
    assert resources['RawSerialPort'] == {None: ({'port': 'foo', 'speed': 115200},)}

    assert 'FakeConsoleDriver' in drivers
    assert drivers['FakeConsoleDriver'] == {'console': ({}, {})}

    assert config == original_config


def test_convert_error():
    with pytest.raises(InvalidConfigError) as excinfo:
        data = load("""
        - {}
        """)
        target_factory._convert_to_named_list(data)
    assert "invalid empty dict as list item" in excinfo.value.msg

    with pytest.raises(InvalidConfigError) as excinfo:
        data = load("""
        - "error"
        """)
        target_factory._convert_to_named_list(data)
    assert "invalid list item type <class 'str'> (should be dict)" in excinfo.value.msg

    with pytest.raises(InvalidConfigError) as excinfo:
        data = load("""
        - name: "bar"
          extra: "baz"
        """)
        target_factory._convert_to_named_list(data)
    assert "missing 'cls' key in OrderedDict(" in excinfo.value.msg

    with pytest.raises(InvalidConfigError) as excinfo:
        data = load("""
        - one:
        - two: {}
        """)
        target_factory._convert_to_named_list(data)
    assert "invalid list item, add empty dict for no arguments" in excinfo.value.msg


def test_resource_param_error():
    with pytest.raises(InvalidConfigError) as excinfo:
        target_factory.make_resource(
            None, 'NetworkSerialPort', 'serial', {'port': None})
    assert "failed to create" in excinfo.value.msg


def test_driver_param_error():
    with pytest.raises(InvalidConfigError) as excinfo:
        target_factory.make_driver(
            None, 'QEMUDriver', 'qemu', {'cpu': 'arm'})
    assert "failed to create" in excinfo.value.msg


def test_resource_class_error():
    with pytest.raises(InvalidConfigError) as excinfo:
        target_factory.make_resource(
            None, 'UnknownResource', None, {})
    assert "unknown resource class" in excinfo.value.msg


def test_driver_class_error():
    with pytest.raises(InvalidConfigError) as excinfo:
        target_factory.make_driver(
            None, 'UnknownDriver', None, {})
    assert "unknown driver class" in excinfo.value.msg


def test_register_same_driver():

    @attr.s
    class SameDriver(Driver):
        pass

    with pytest.raises(RegistrationError) as excinfo:
        target_factory.reg_driver(SameDriver)
        target_factory.reg_driver(SameDriver)
    assert "driver with name" in excinfo.value.msg

def test_register_same_resource():

    @attr.s
    class SameResource(Resource):
        pass

    with pytest.raises(RegistrationError) as excinfo:
        target_factory.reg_driver(SameResource)
        target_factory.reg_driver(SameResource)
    assert "driver with name" in excinfo.value.msg



"""The QEMUDriver implements a driver to use a QEMU target"""
import atexit
import select
import shlex
import shutil
import socket
import subprocess
import re
import tempfile
import time

import attr
from pexpect import TIMEOUT

from ..factory import target_factory
from ..protocol import PowerProtocol, ConsoleProtocol
from ..step import step
from .common import Driver
from .consoleexpectmixin import ConsoleExpectMixin
from ..util.qmp import QMPMonitor, QMPError
from .exception import ExecutionError


@target_factory.reg_driver
@attr.s(eq=False)
class QEMUDriver(ConsoleExpectMixin, Driver, PowerProtocol, ConsoleProtocol):
    """
    The QEMUDriver implements an interface to start targets as qemu instances.

    The kernel, flash, rootfs and dtb arguments refer to images and paths
    declared in the environment configuration.

    Args:
        qemu_bin (str): reference to the tools key for the QEMU binary
        machine (str): QEMU machine type
        cpu (str): QEMU cpu type
        memory (str): QEMU memory size (ends with M or G)
        extra_args (str): optional, extra QEMU arguments passed directly to the QEMU binary
        boot_args (str): optional, additional kernel boot argument
        kernel (str): optional, reference to the images key for the kernel
        disk (str): optional, reference to the images key for the disk image
        disk_opts (str): optional, additional QEMU disk options
        flash (str): optional, reference to the images key for the flash image
        rootfs (str): optional, reference to the paths key for use as the virtio-9p filesystem
        dtb (str): optional, reference to the image key for the device tree
        bios (str): optional, reference to the image key for the bios image
        display (str, default="none"): optional, display output to enable; must be one of:
            none: Do not create a display device
            fb-headless: Create a headless framebuffer device
            egl-headless: Create a headless GPU-backed graphics card. Requires host support
            qemu-default: Don't override QEMU default settings
        nic (str): optional, configuration string to pass to QEMU to create a network interface
    """
    qemu_bin = attr.ib(validator=attr.validators.instance_of(str))
    machine = attr.ib(validator=attr.validators.instance_of(str))
    cpu = attr.ib(validator=attr.validators.instance_of(str))
    memory = attr.ib(validator=attr.validators.instance_of(str))
    extra_args = attr.ib(
        default='',
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    boot_args = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    kernel = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    disk = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    disk_opts = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    rootfs = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    dtb = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    flash = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    bios = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    display = attr.ib(
        default="none",
        validator=attr.validators.optional(attr.validators.and_(
            attr.validators.instance_of(str),
            attr.validators.in_(
                ["none", "fb-headless", "egl-headless", "qemu-default"]
            ),
        ))
    )
    nic = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.status = 0
        self.txdelay = None
        self.txchunk = None
        self._child = None
        self._tempdir = None
        self._socket = None
        self._clientsocket = None
        self._forwarded_ports = {}
        atexit.register(self._atexit)

    def _atexit(self):
        if not self._child:
            return
        self._child.terminate()
        try:
            self._child.communicate(timeout=1)
        except subprocess.TimeoutExpired:
            self._child.kill()
            self._child.communicate(timeout=1)

    def get_qemu_version(self, qemu_bin):
        p = subprocess.run([qemu_bin, "-version"], stdout=subprocess.PIPE, encoding="utf-8")
        if p.returncode != 0:
            raise ExecutionError(f"Unable to get QEMU version. QEMU exited with: {p.returncode}")

        m = re.search(r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<micro>\d+)', p.stdout.splitlines()[0])
        if m is None:
            raise ExecutionError(f"Unable to find QEMU version in: {p.stdout.splitlines()[0]}")

        return (int(m.group('major')), int(m.group('minor')), int(m.group('micro')))

    def get_qemu_base_args(self):
        """Returns the base command line used for Qemu without the options
        related to QMP. These options can be used to start an interactive
        Qemu manually for debugging tests
        """
        cmd = []

        qemu_bin = self.target.env.config.get_tool(self.qemu_bin)
        if qemu_bin is None:
            raise KeyError(
                "QEMU Binary Path not configured in tools configuration key")
        cmd = [qemu_bin]

        qemu_version = self.get_qemu_version(qemu_bin)

        boot_args = []

        if self.kernel is not None:
            cmd.append("-kernel")
            cmd.append(
                self.target.env.config.get_image_path(self.kernel))
        if self.disk is not None:
            disk_path = self.target.env.config.get_image_path(self.disk)
            disk_format = "raw"
            if disk_path.endswith(".qcow2"):
                disk_format = "qcow2"
            disk_opts = ""
            if self.disk_opts:
                disk_opts = f",{self.disk_opts}"
            if self.machine == "vexpress-a9":
                cmd.append("-drive")
                cmd.append(
                    f"if=sd,format={disk_format},file={disk_path},id=mmc0{disk_opts}")
                boot_args.append("root=/dev/mmcblk0p1 rootfstype=ext4 rootwait")
            elif self.machine in ["pc", "q35", "virt"]:
                cmd.append("-drive")
                cmd.append(
                    f"if=virtio,format={disk_format},file={disk_path}{disk_opts}")
                boot_args.append("root=/dev/vda rootwait")
            else:
                raise NotImplementedError(
                    f"QEMU disk image support not implemented for machine '{self.machine}'"
                )
        if self.rootfs is not None:
            cmd.append("-fsdev")
            cmd.append(
                f"local,id=rootfs,security_model=none,path={self.target.env.config.get_path(self.rootfs)}")  # pylint: disable=line-too-long
            cmd.append("-device")
            cmd.append(
                "virtio-9p-device,fsdev=rootfs,mount_tag=/dev/root")
            boot_args.append("root=/dev/root rootfstype=9p rootflags=trans=virtio")
        if self.dtb is not None:
            cmd.append("-dtb")
            cmd.append(self.target.env.config.get_image_path(self.dtb))
        if self.flash is not None:
            cmd.append("-drive")
            cmd.append(
                f"if=pflash,format=raw,file={self.target.env.config.get_image_path(self.flash)},id=nor0")  # pylint: disable=line-too-long
        if self.bios is not None:
            cmd.append("-bios")
            cmd.append(
                self.target.env.config.get_image_path(self.bios))

        if "-append" in shlex.split(self.extra_args):
            raise ExecutionError("-append in extra_args not allowed, use boot_args instead")

        cmd.extend(shlex.split(self.extra_args))
        cmd.append("-machine")
        cmd.append(self.machine)
        cmd.append("-cpu")
        cmd.append(self.cpu)
        cmd.append("-m")
        cmd.append(self.memory)
        if self.display == "none":
            cmd.append("-nographic")
        elif self.display == "fb-headless":
            cmd.append("-display")
            cmd.append("none")
        elif self.display == "egl-headless":
            if qemu_version >= (6, 1, 0):
                cmd.append("-device")
                cmd.append("virtio-vga-gl")
            else:
                cmd.append("-vga")
                cmd.append("virtio")
            cmd.append("-display")
            cmd.append("egl-headless")
        elif self.display != "qemu-default":
            raise ExecutionError(f"Unknown display '{self.display}'")

        if self.nic:
            cmd.append("-nic")
            cmd.append(self.nic)

        if self.boot_args is not None:
            boot_args.append(self.boot_args)
        if self.kernel is not None and boot_args:
            cmd.append("-append")
            cmd.append(" ".join(boot_args))

        return cmd

    def on_activate(self):
        self._tempdir = tempfile.mkdtemp(prefix="labgrid-qemu-tmp-")
        sockpath = f"{self._tempdir}/serialrw"
        self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._socket.bind(sockpath)
        self._socket.listen(0)

        self._cmd = self.get_qemu_base_args()

        self._cmd.append("-S")
        self._cmd.append("-qmp")
        self._cmd.append("stdio")

        self._cmd.append("-chardev")
        self._cmd.append(f"socket,id=serialsocket,path={sockpath}")
        self._cmd.append("-serial")
        self._cmd.append("chardev:serialsocket")

    def on_deactivate(self):
        if self.status:
            self.off()
        if self._clientsocket:
            self._clientsocket.close()
            self._clientsocket = None
        self._socket.close()
        self._socket = None
        shutil.rmtree(self._tempdir)

    @step()
    def on(self):
        """Start the QEMU subprocess, accept the unix socket connection and
        afterwards start the emulator using a QMP Command"""
        if self.status:
            return
        self.logger.debug("Starting with: %s", self._cmd)
        self._child = subprocess.Popen(
            self._cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        # prepare for timeout handing
        self._clientsocket, address = self._socket.accept()
        self._clientsocket.setblocking(0)
        self.logger.debug("new connection from %s", address)

        try:
            self.qmp = QMPMonitor(self._child.stdout, self._child.stdin)
        except QMPError as exc:
            if self._child.poll() is not None:
                _, err = self._child.communicate()
                raise IOError(
                    f"QEMU error: {err} (exitcode={self._child.returncode})"
                ) from exc
            raise

        self.status = 1

        # Restore port forwards
        for v in self._forwarded_ports.values():
            self._add_port_forward(*v)

        self.monitor_command("cont")

    @step()
    def off(self):
        """Stop the emulator using a monitor command and await the exitcode"""
        if not self.status:
            return
        self.monitor_command('quit')
        if self._child.wait() != 0:
            self._child.communicate()
            raise IOError
        self._child = None
        self.status = 0

    def cycle(self):
        """Cycle the emulator by restarting it"""
        self.off()
        self.on()

    @step(result=True, args=['command', 'arguments'])
    def monitor_command(self, command, arguments={}):
        """Execute a monitor_command via the QMP"""
        if not self.status:
            raise ExecutionError(
                "Can't use monitor command on non-running target")
        return self.qmp.execute(command, arguments)

    def _add_port_forward(self, proto, local_address, local_port, remote_address, remote_port, netdev):
        command = ("hostfwd_add", netdev, f"{proto}:{local_address}:{local_port}-{remote_address}:{remote_port}")
        command = filter(None, command)
        self.monitor_command(
            "human-monitor-command",
            {"command-line": " ".join(command)},
        )

    def add_port_forward(self, proto, local_address, local_port, remote_address, remote_port, netdev=""):
        self._add_port_forward(proto, local_address, local_port, remote_address, remote_port, netdev)
        self._forwarded_ports[(proto, local_address, local_port, netdev)] = (proto, local_address, local_port, remote_address, remote_port, netdev)

    def remove_port_forward(self, proto, local_address, local_port, netdev=""):
        del self._forwarded_ports[(proto, local_address, local_port, netdev)]
        command = ("hostfwd_remove", netdev, f"{proto}:{local_address}:{local_port}")
        command = filter(None, command)
        self.monitor_command(
            "human-monitor-command",
            {"command-line": " ".join(command)},
        )

    def _read(self, size=1, timeout=10, max_size=None):
        ready, _, _ = select.select([self._clientsocket], [], [], timeout)
        if ready:
            # Collect some more data
            time.sleep(0.01)
            # Always read a page, regardless of size
            size = 4096
            size = min(max_size, size) if max_size else size
            res = self._clientsocket.recv(size)
        else:
            raise TIMEOUT(f"Timeout of {timeout:.2f} seconds exceeded")
        return res

    def _write(self, data):
        return self._clientsocket.send(data)

    def __str__(self):
        return f"QemuDriver({self.target.name})"



import base64
import hashlib
import json
import os.path
import socket
import subprocess
import traceback
import logging

from .ssh import get_ssh_connect_timeout

def b2s(b):
    return base64.b85encode(b).decode('ascii')

def s2b(s):
    return base64.b85decode(s.encode('ascii'))

class AgentError(Exception):
    pass

class AgentException(Exception):
    pass

class MethodProxy:
    def __init__(self, wrapper, name):
        self.wrapper = wrapper
        self.name = name

    def __call__(self, *args, **kwargs):
        return self.wrapper.call(self.name, *args, **kwargs)

class ModuleProxy:
    def __init__(self, wrapper, name):
        self.wrapper = wrapper
        self.name = name

    def __getattr__(self, name):
        return MethodProxy(self.wrapper, f'{self.name}.{name}')

class AgentWrapper:
    def __init__(self, host=None):
        self.agent = None
        self.loaded = {}
        self.logger = logging.getLogger(f"ResourceExport({host})")
        self.fdpass = None

        agent = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            'agent.py')
        agent_prefix = os.environ.get("LG_AGENT_PREFIX", "")
        if host:
            # copy agent.py and run via ssh
            with open(agent, 'rb') as agent_fd:
                agent_data = agent_fd.read()
            agent_hash = hashlib.sha256(agent_data).hexdigest()
            agent_remote = os.path.join(agent_prefix, f'.labgrid_agent_{agent_hash}.py')
            connect_timeout = get_ssh_connect_timeout()
            ssh_opts = f'ssh -x -o ConnectTimeout={connect_timeout} -o PasswordAuthentication=no'.split()
            subprocess.check_call(
                ['rsync', '-e', ' '.join(ssh_opts), '-tq', agent,
                 f'{host}:{agent_remote}'],
            )
            self.agent = subprocess.Popen(
                ssh_opts + [host, '--', 'python3', agent_remote],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                start_new_session=True,
            )
        else:
            # run locally
            self.fdpass, remote_fdpass = socket.socketpair()
            with remote_fdpass:
                env = os.environ.copy()
                env["LG_FDPASS"] = str(remote_fdpass.fileno())
                self.agent = subprocess.Popen(
                    ['python3', agent],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    env=env,
                    start_new_session=True,
                    pass_fds=(remote_fdpass.fileno(),),
                )

    def __del__(self):
        self.close()

    def __getattr__(self, name):
        return MethodProxy(self, name)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def call(self, method, *args, **kwargs):
        request = {
            'method': method,
            'args': args,
            'kwargs': kwargs,
            }
        request = json.dumps(request)
        request = request.encode('ASCII')
        self.agent.stdin.write(request+b'\n')
        self.agent.stdin.flush()
        response = self.agent.stdout.readline()
        response = response.decode('ASCII')
        response = json.loads(response)
        if 'result' in response:
            if response.get('fdpass'):
                _, fds, _, _ = socket.recv_fds(self.fdpass, 1, 1)
                return (response['result'], fds[0])
            else:
                return response['result']
        elif 'exception' in response:
            e = response['exception']
            # work around BaseException repr change
            # https://bugs.python.org/issue30399
            if e[-2:] == ',)':
                e = e[:-2] + ')'
            self.logger.debug("Traceback from agent (most recent call last) for %s:", e)
            for line in ''.join(traceback.format_list(response['tb'])).splitlines():
                self.logger.debug(line)
            raise AgentException(e)
        elif 'error' in response:
            self.agent.wait()
            self.agent.communicate()
            self.agent = None
            raise AgentError(response['error'])

        raise AgentError(f"unknown response from agent: {response}")

    def load(self, name, path=None):
        if name in self.loaded:
            return self.loaded[name]

        if path is None:
            path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'agents')

        filename = os.path.join(path, f'{name}.py')
        with open(filename, 'r') as source_fd:
            source = source_fd.read()

        self.call('load', name, source)

        proxy = ModuleProxy(self, name)
        self.loaded[name] = proxy
        return proxy

    def close(self):
        if self.agent is None:
            return
        request = {
            'close': True,
            }
        request = json.dumps(request)
        request = request.encode('ASCII')
        self.agent.stdin.write(request+b'\n')
        self.agent.stdin.flush()
        self.agent.wait()
        self.agent.communicate()
        self.agent = None
